"""
The ArenaCmdInterpreter is a simple line-oriented command interpreter that
allows to inspect library/program state. It looks at :envvar:`ENABLE_INTERPRETER`
to enable the interpreter.

The :class:`.ArenaCmdInterpreter` receives a :class:`.Scene` instance and provides commands
to inspect attributes and execute functions (callables) given to the constructor.

The commands available are:

* show: displays attributes
* info: executes scene functions that output information
* help: displays the commands available
* exit: terminates the program

"""

import asyncio
import cmd
import json
import os
import threading
import time
from datetime import date, datetime

from ..env import ENABLE_INTERPRETER, _get_env


class ArenaCmdInterpreter(cmd.Cmd):
    intro = 'Welcome to the arena-py console. Type help or ? to list available commands.\n'
    prompt = '# '
    file = None

    def __serialize_obj(self, obj):
        if isinstance(obj, (datetime, date)):
            return str(obj)
        if hasattr(obj, '__dict__'):
            return {k: v for k, v in vars(obj).items() if k.startswith("_") == False} # ignore private members
        raise TypeError("Type not serializable")

    def __init__(self, scene, show_attrs=('config_data', 'scene', 'users', 'all_objects', 'msg_io'), get_callables=('persisted_objs', 'persisted_scene_option', 'writable_scenes', 'user_list'), start_cmd_event=None):
        self.enable_interp = _get_env(ENABLE_INTERPRETER).lower() in ('true', '1', 't')
        if not self.enable_interp: return
        super().__init__(completekey='tab')
        self._scene = scene
        self._show_attrs = show_attrs
        self._get_callables = get_callables

        # start interpreter thread
        t = threading.Thread(name='interpreter_thread', target=self.__cmd_loop_thread, args=(start_cmd_event,))
        t.start()

    def __cmd_loop_thread(self, start_cmd_event):
        if start_cmd_event: start_cmd_event.wait(5) # try to start cmd last; wait on event with timeout
        self.cmdloop()

    def do_show(self, arg):
        if arg not in self._show_attrs:
            self.help_show()
            return
        try:
            obj = getattr(self._scene, arg)
        except AttributeError:
            print(f"Could not find attr {arg}")
            return

        print(json.dumps(obj, indent=4, sort_keys=True, default=self.__serialize_obj))

    def help_show(self):
        print(f"Display scene attributes: {[i for i in self._show_attrs]}")

    def do_get(self, arg):
        if arg not in self._get_callables:
            self.help_get()
            return
        try:
            scene_get = getattr(self._scene, f"get_{arg}")
        except AttributeError:
            print(f"Could not find attr get_{arg}")
            return

        try:
            if callable(scene_get):
                print(scene_get())
        except Exception as e:
            print(e)

    def help_get(self):
        print(f"Scene get_* methods: {[i for i in self._get_callables]}. E.g:\n")
        print("\tget persisted_objs => returns all persisted objects in the scene (by executing scene.get_persisted_objs)\n")

    def do_exit(self, arg):
        answer = ""
        while answer not in ["y", "n"]:
            answer = input("This will terminate the ARENA program. Are you sure [Y/N]? ").lower()
        if answer == "y":
            print("Exiting...")
            self._scene.exit(0)

        return True

    def help_exit(self):
        print("Exit program.")



from deprecated import deprecated

from .arena_object import Object


@deprecated("Cube (=Box) Geometry (deprecated); Supported for Legacy reasons; Please use Box in new scenes")
class Cube(Object):
    """
    Cube object class to manage its properties in the ARENA: Cube (=Box) Geometry (**deprecated**); Supported for Legacy reasons; Please use Box in new scenes

    :param float depth: *depth* Defaults to '1' (deprecated)
    :param float height: *height* Defaults to '1' (deprecated)
    :param int segmentsDepth: *segments depth* Defaults to '1' (deprecated)
    :param int segmentsHeight: *segments height* Defaults to '1' (deprecated)
    :param int segmentsWidth: *segments width* Defaults to '1' (deprecated)
    :param float width: *width* Defaults to '1' (deprecated)
    """
    object_type = "cube"

    def __init__(self, **kwargs):
        super().__init__(object_type=Cube.object_type, **kwargs)



from .arena_object import Object

class Cylinder(Object):
    """
    Cylinder object class to manage its properties in the ARENA: Cylinder Geometry.

    :param float height: height Defaults to '1' (optional)
    :param bool openEnded: open ended Defaults to 'False' (optional)
    :param float radius: radius Defaults to '1' (optional)
    :param int segmentsHeight: segments height Defaults to '18' (optional)
    :param int segmentsRadial: segments radial Defaults to '36' (optional)
    :param float thetaLength: theta length Defaults to '360' (optional)
    :param float thetaStart: theta start Defaults to '0' (optional)
    """
    object_type = "cylinder"

    def __init__(self, **kwargs):
        super().__init__(object_type=Cylinder.object_type, **kwargs)



from .arena_object import Object


class GaussianSplatting(Object):
    """
    GaussianSplatting object class to manage its properties in the ARENA: Load a 3D Gaussian Splat for Real-Time Radiance Field Rendering. More information: <https://github.com/quadjr/aframe-gaussian-splatting> A-Frame Gaussian Splatting. See guidance to store paths under <https://docs.arenaxr.org/content/interface/filestore.html> ARENA File Store, CDN, or DropBox.

    :param str cutoutEntity: Selector to a box primitive that uses scale and position to define the bounds of splat points to render. Defaults to '' (optional)
    :param float pixelRatio: Pixel ratio for rendering. Reducing the value decreases the resolution and improves performance. If a negative value is set, the device's native value will be applied. Defaults to '1.0' (optional)
    :param str src: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    :param float xrPixelRatio: Same as pixelRatio. Applied to XR devices. Defaults to '0.5' (optional)
    """
    object_type = "gaussian_splatting"

    def __init__(self, **kwargs):
        super().__init__(object_type=GaussianSplatting.object_type, **kwargs)



import uuid

from ..attributes import DataEvent
from ..base_object import *


class Event(BaseObject):
    """
    Event class. Wrapper around JSON for events.
    """
    def __init__(self, **kwargs):
        # "object_id" is required in kwargs, defaulted to random uuid4
        object_id = kwargs.get("object_id", str(uuid.uuid4()))
        if "object_id" in kwargs: del kwargs["object_id"]

        # make "action" "clientEvent" by default and remove "action" from kwargs
        action = kwargs.get("action", "clientEvent")
        if "action" in kwargs: del kwargs["action"]

        # make "type" "mousedown" by default and remove "type" from kwargs
        _type = kwargs.get("type", "mousedown")
        if "type" in kwargs: del kwargs["type"]

        kwargs = kwargs.get("data", kwargs)
        data = DataEvent(**kwargs)
        super().__init__(
                object_id=object_id,
                action=action,
                type=_type,
                data=data
            )

    # TODO (mwfarb): We should standardize this json() transform into BaseObject from Object/Event/Program
    def json(self, **kwargs):
        json_payload = vars(self).copy()
        json_payload.update(kwargs)

        data = vars(json_payload["data"])
        json_data = {}
        for k, v in data.items():
            if v is None:
                json_data[k] = v

            # rotation should be in quaternions
            if "rotation" == k:
                rot = data["rotation"]
                # always publish quaternions on wire format to avoid persist euler->quat merges
                json_data["rotation"] = rot.quaternion

            else:
                json_data[k] = v

        json_payload["data"] = json_data

        return self.json_encode(json_payload)



from .arena_object import Object

class Sphere(Object):
    """
    Sphere object class to manage its properties in the ARENA: Sphere Geometry.

    :param float phiLength: phi length Defaults to '360' (optional)
    :param float phiStart: phi start Defaults to '0' (optional)
    :param float radius: radius Defaults to '1' (optional)
    :param int segmentsHeight: segments height Defaults to '18' (optional)
    :param int segmentsWidth: segments width Defaults to '36' (optional)
    :param float thetaLength: theta length Defaults to '180' (optional)
    :param float thetaStart: theta start Defaults to '0' (optional)
    """
    object_type = "sphere"

    def __init__(self, **kwargs):
        super().__init__(object_type=Sphere.object_type, **kwargs)



from .attribute import Attribute

class Material(Attribute):
    """
    Material attribute class to manage its properties in the ARENA: The material properties of the object's surface. More properties at <https://aframe.io/docs/1.5.0/components/material.html> A-Frame Material.
    Usage: `material=Material(...)`

    :param float alphaTest: Alpha test threshold for transparency. Defaults to '0' (optional)
    :param float anisotropy: The anisotropic filtering sample rate to use for the textures. A value of 0 means the default value will be used, see renderer. Defaults to '0' (optional)
    :param str blending: The blending mode for the material's RGB and Alpha sent to the WebGLRenderer. Allows [none, normal, additive, subtractive, multiply] Defaults to 'normal' (optional)
    :param str color: Base diffuse color. Defaults to '#ffffff' (optional)
    :param str combine: How the environment map mixes with the material. Requires shader: phong. Allows [mix, add, multiply] Defaults to 'mix' (optional)
    :param bool depthTest: Whether depth testing is enabled when rendering the material. Defaults to 'True' (optional)
    :param bool depthWrite: Render when depth test succeeds. Defaults to 'True' (optional)
    :param bool dithering: Whether material is dithered with noise. Removes banding from gradients like ones produced by lighting. Defaults to 'True' (optional)
    :param str emissive: The color of the emissive lighting component. Used to make objects produce light even without other lighting in the scene. Requires shader: standard or phong Defaults to '#000000' (optional)
    :param float emissiveIntensity: Intensity of the emissive lighting component. Requires shader: standard or phong Defaults to '1' (optional)
    :param bool flatShading: Use THREE.FlatShading rather than THREE.StandardShading. Defaults to 'False' (optional)
    :param bool fog: Whether or not material is affected by fog. Defaults to 'True' (optional)
    :param int height: Height of video (in pixels), if defining a video texture. Requires shader: standard or flat. Defaults to '256' (optional)
    :param float metalness: How metallic the material is from 0 to 1. Requires shader: standard. Defaults to '0' (optional)
    :param bool npot: Use settings for non-power-of-two (NPOT) texture. Defaults to 'False' (optional)
    :param dict offset: Texture offset to be used. Defaults to '{'x': 0, 'y': 0}' (optional)
    :param float opacity: Extent of transparency. If the transparent property is not true, then the material will remain opaque and opacity will only affect color. Defaults to '1' (optional)
    :param float reflectivity: How much the environment map affects the surface. Requires shader: phong. Defaults to '0.9' (optional)
    :param bool refract: Whether the defined envMap should refract. Requires shader: phong. Defaults to 'False' (optional)
    :param float refractionRatio: 1/refractive index of the material. Requires shader: phong. Defaults to '0.98' (optional)
    :param dict repeat: How many times a texture (defined by src) repeats in the X and Y direction. Defaults to '{'x': 1, 'y': 1}' (optional)
    :param float roughness: How rough the material is from 0 to 1. A rougher material will scatter reflected light in more directions than a smooth material. Requires shader: standard. Defaults to '0' (optional)
    :param str shader: Which material to use. Defaults to the standard material. Can be set to the flat material or to a registered custom shader material. Allows [flat, standard, phong] Defaults to 'standard' (optional)
    :param float shininess: How shiny the specular highlight is; a higher value gives a sharper highlight. Requires shader: phong. Defaults to '30' (optional)
    :param str side: Which sides of the mesh to render. Allows [front, back, double] Defaults to 'front' (optional)
    :param str specular: This defines how shiny the material is and the color of its shine. Requires shader: phong. Defaults to '#111111' (optional)
    :param str src: URI, relative or full path of an image/video file. e.g. 'store/users/wiselab/images/360falls.mp4'. (optional)
    :param bool toneMapped: Whether to ignore toneMapping, set to false you are using renderer.toneMapping and an element should appear to emit light. Requires shader: flat. Defaults to 'True' (optional)
    :param bool transparent: Whether material is transparent. Transparent entities are rendered after non-transparent entities. Defaults to 'False' (optional)
    :param bool vertexColorsEnabled: Whether to use vertex or face colors to shade the material. Defaults to 'False' (optional)
    :param bool visible: Whether material is visible. Raycasters will ignore invisible materials. Defaults to 'True' (optional)
    :param int width: Width of video (in pixels), if defining a video texture. Requires shader: standard or flat. Defaults to '512' (optional)
    :param bool wireframe: Whether to render just the geometry edges. Defaults to 'False' (optional)
    :param int wireframeLinewidth: Width in px of the rendered line. Defaults to '2' (optional)
    """
    def __init__(self, **kwargs):
        if "opacity" in kwargs:
            # kwargs["transparent"] = True # need to be transparent to be opaque
            if kwargs["opacity"] > 1.0: # keep opacity between 0.0 and 1.0
                kwargs["opacity"] = float(kwargs["opacity"]) / 100
            kwargs["opacity"] = max(0.0, kwargs["opacity"])
            kwargs["opacity"] = min(kwargs["opacity"], 1.0)

        super().__init__(**kwargs)



from .utils import Utils
from .cmd_interpreter import ArenaCmdInterpreter
from .arena_telemetry import ArenaTelemetry
from .program_info import ProgramRunInfo, QueueStats
from .mqtt import topic_matches_sub



from .arena_object import Object

class Roundedbox(Object):
    """
    Roundedbox object class to manage its properties in the ARENA: Rounded Box Geometry.

    :param float depth: depth Defaults to '1' (optional)
    :param float height: height Defaults to '1' (optional)
    :param float radius: radius of edge Defaults to '0.15' (optional)
    :param int radiusSegments: segments radius Defaults to '10' (optional)
    :param float width: width Defaults to '1' (optional)
    """
    object_type = "roundedbox"

    def __init__(self, **kwargs):
        super().__init__(object_type=Roundedbox.object_type, **kwargs)



from .attribute import Attribute

class JitsiVideo(Attribute):
    """
    JitsiVideo attribute class to manage its properties in the ARENA: Apply a jitsi video source to the geometry.
    Usage: `jitsi_video=JitsiVideo(...)`

    :param str displayName: ARENA or Jitsi display name of the video source; Will be ignored if jitsiId is given. (* change requires reload* ). (optional)
    :param str jitsiId: JitsiId of the video source; If defined will override displayName. (optional)
    """
    def __init__(self, jitsiId=None, displayName=None):
        # Since jitsiId overrides displayName, do not set None arguments
        kwargs = {}
        if jitsiId is not None:
            kwargs['jitsiId'] = jitsiId
        if displayName is not None:
            kwargs['displayName'] = displayName
        super().__init__(**kwargs)



from .arena_object import Object

class ArenauiPrompt(Object):
    """
    ArenauiPrompt object class to manage its properties in the ARENA: ARENAUI element which displays prompt with button actions.

    :param list[str] buttons: Buttons Defaults to '['Confirm', 'Cancel']' (optional)
    :param str description: Description Defaults to 'This is a prompt. Please confirm or cancel.' (optional)
    :param str font: Font to use for button text. Allows [Roboto, Roboto-Mono] Defaults to 'Roboto' (optional)
    :param str materialSides: Which sides display the rendered UI material Allows [both, front] Defaults to 'both' (optional)
    :param str theme: Color Theme Allows [light, dark] Defaults to 'light' (optional)
    :param str title: Title Defaults to 'Prompt' (optional)
    :param float width: Override width Defaults to '1.5' (optional)
    """
    object_type = "arenaui-prompt"

    def __init__(self, **kwargs):
        super().__init__(object_type=ArenauiPrompt.object_type, **kwargs)


class Prompt(ArenauiPrompt):
    """
    Alternate name for ArenauiPrompt.
    """



from .attribute import Attribute


class Shadow(Attribute):
    """
    Shadow attribute class to manage its properties in the ARENA: The shadow component enables shadows for an entity and its children. Adding the shadow component alone is not enough to display shadows in your scene. We must have at least one light with castShadow: true enabled.
    Usage: `shadow=Shadow(...)`

    :param bool cast: Whether the entity casts shadows onto the surrounding scene. Defaults to 'False' (optional)
    :param bool receive: Whether the entity receives shadows from the surrounding scene. Defaults to 'False' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from deprecated import deprecated

from ..utils import *
from .attribute import Attribute
from .position import Position


class DataEvent(Attribute):
    def __init__(self, **kwargs):
        data = {}
        data = DataEvent.update_data(data, kwargs)
        super().__init__(**data)

    @classmethod
    def update_data(cls, data, new_data):
        new_data = new_data.get("data", new_data)
        for k, v in new_data.items():
            # allow user to input tuples, lists, dicts, etc for specific Attributes.
            # everything gets converted to corresponding attribute
            if (k == "originPosition" or k == "targetPosition") and not isinstance(v, Position):
                if isinstance(v, (list, tuple)):
                    data[k] = Position(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Position(**v)
                else:
                    data[k] = v

            elif isinstance(v, Attribute):
                data[k] = v

            else:
                try:
                    # unknown attribute
                    data[k] = Attribute(**v)
                except:
                    data[k] = v

        return data

    @property
    @deprecated("DEPRECATED: data.source is deprecated for clientEvent, use data.target instead.")
    def source(self):
        return None

    @source.setter
    @deprecated("DEPRECATED: data.source is deprecated for clientEvent, use data.target instead.")
    def source(self, value):
        return

    @property
    @deprecated("DEPRECATED: data.clickPos is deprecated for clientEvent, use data.originPosition instead.")
    def clickPos(self):
        return None

    @clickPos.setter
    @deprecated("DEPRECATED: data.clickPos is deprecated for clientEvent, use data.originPosition instead.")
    def clickPos(self, value):
        return

    @property
    @deprecated("DEPRECATED: data.position is deprecated for clientEvent, use data.targetPosition instead.")
    def position(self):
        return None

    @position.setter
    @deprecated("DEPRECATED: data.position is deprecated for clientEvent, use data.targetPosition instead.")
    def position(self, value):
        return



from .arena_object import Object

class Octahedron(Object):
    """
    Octahedron object class to manage its properties in the ARENA: Octahedron Geometry.

    :param int detail: detail Defaults to '0' (optional)
    :param float radius: radius Defaults to '1' (optional)
    """
    object_type = "octahedron"

    def __init__(self, **kwargs):
        super().__init__(object_type=Octahedron.object_type, **kwargs)



from .attribute import Attribute


class ClickListener(Attribute):
    """
    ClickListener attribute class to manage its properties in the ARENA: Object will listen for mouse events like clicks.
    Usage: `click_listener=ClickListener(...)` or `clickable=True`

    :param bool bubble: Set false to prevent click events from bubbling up to parent objects. See <https://developer.mozilla.org/en-US/docs/Learn/JavaScript/Building_blocks/Events#event_bubbling> Event Bubbling. Defaults to 'True' (optional)
    :param bool enabled: Publish events, set false to disable. Defaults to 'True' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class DynamicBody(Attribute):
    """
    DynamicBody attribute class to manage its properties in the ARENA: A freely-moving object. Dynamic bodies have mass, collide with other objects, bounce or slow during collisions, and fall if gravity is enabled. More properties at <https://github.com/c-frame/aframe-physics-system/blob/master/CannonDriver.md> A-Frame Physics System.
    Usage: `dynamic_body=DynamicBody(...)`

    :param float angularDamping: Resistance to rotation. Defaults to '0.01' (optional)
    :param str cylinderAxis: Override default axis of bounding cylinder. Requires shape:cylinder. Allows [x, y, z] Defaults to 'y' (optional)
    :param float linearDamping: Resistance to movement. Defaults to '0.01' (optional)
    :param float mass: Simulated mass of the object, > 0. Defaults to '5' (optional)
    :param str shape: Body components will attempt to find an appropriate CANNON.js shape to fit your model. When defining an object you may choose a shape or leave the default, auto. Select a shape carefully, as there are performance implications with different choices. Allows [auto, box, cylinder, sphere, hull, none] Defaults to 'auto' (optional)
    :param float sphereRadius: Override default radius of bounding sphere. Requires shape:sphere. NaN by default. (optional)
    :param str type: Define the result of collisions. Dynamic can be moved, Static cannot be moved. Allows [dynamic, static] Defaults to 'dynamic' (optional)
    """

    def __init__(self, type="static", **kwargs):
        _type = type
        if _type != "static" and _type != "dynamic":
            _type = "none"
        super().__init__(type=_type, **kwargs)


class Physics(DynamicBody):
    """
    Alternate name for DynamicBody.
    Usage: `physics=Physics(...)`
    """



from ..utils import *
from .attribute import Attribute
from .color import Color
from .dynamic_body import Physics
from .goto_url import GotoUrl
from .jitsi_video import JitsiVideo
from .material import Material
from .position import Position
from .rotation import Rotation
from .scale import Scale
from .translate import ATTRIBUTE_CLASS_TRANSLATION, KEYWORD_ATTRIBUTE_TRANSLATION
from .video_control import VideoControl


class Data(Attribute):
    """
    Data attribute class to manage its properties in the ARENA: Wraps all attributes in JSON.
    Usage: `data=Data(...)`

    :param dict animation: Animate and tween values. More properties at <https://aframe.io/docs/1.5.0/components/animation.html> A-Frame Animation component. Easing properties are detailed at <https://easings.net> easings.net. (optional)
    :param dict animation_mixer: A list of available animations can usually be found by inspecting the model file or its documentation. All animations will play by default. To play only a specific set of animations, use wildcards: animation-mixer='clip: run_*'. More properties at <https://github.com/n5ro/aframe-extras/tree/master/src/loaders#animation> A-Frame Extras Animation. (optional)
    :param dict armarker: A location marker (such as an AprilTag, a lightAnchor, or an UWB tag), used to anchor scenes, or scene objects, in the real world. (optional)
    :param dict attribution: Attribution Component. Saves attribution data in any entity. (optional)
    :param dict blip: When the object is created or deleted, it will animate in/out of the scene instead of appearing/disappearing instantly. Must have a geometric mesh. (optional)
    :param dict box_collision_listener: Listen for bounding-box collisions with user camera and hands. Must be applied to an object or model with geometric mesh. Collisions are determined by course bounding-box overlaps. (optional)
    :param bool buffer: Transform geometry into a BufferGeometry to reduce memory usage at the cost of being harder to manipulate (geometries only: box, circle, cone, ...). Defaults to 'True' (optional)
    :param dict click_listener: Object will listen for mouse events like clicks. (optional)
    :param str collision_listener: Name of the collision-listener, default can be empty string. Collisions trigger click events. (optional)
    :param str color: ***DEPRECATED**: data.color is **deprecated** for primitive geometry, use data.material.color instead.* (deprecated)
    :param dict dynamic_body: A freely-moving object. Dynamic bodies have mass, collide with other objects, bounce or slow during collisions, and fall if gravity is enabled. More properties at <https://github.com/c-frame/aframe-physics-system/blob/master/CannonDriver.md> A-Frame Physics System. (optional)
    :param dict gltf_model_lod: Simple switch between the default gltf-model and a detailed one when a user camera is within specified distance (optional)
    :param dict gltf_morph: Allows you to target and control a gltf model's morphTargets created in Blender. More properties at <https://github.com/elbobo/aframe-gltf-morph-component> A-Frame GLTF Morph component. (optional)
    :param dict goto_landmark: Teleports user to the landmark with the given name. Requires click-listener. (optional)
    :param dict goto_url: Load new URL when object is clicked. Requires click-listener. (optional)
    :param bool hide_on_enter_ar: Hide object when entering AR. Remove component to *not* hide. (optional)
    :param bool hide_on_enter_vr: Hide object when entering VR. Remove component to *not* hide. (optional)
    :param dict impulse: Apply an impulse to an object to set it in motion. This happens in conjunction with an event. Requires click-listener and physics. (optional)
    :param dict jitsi_video: Apply a jitsi video source to the geometry. (optional)
    :param dict landmark: Define entities as a landmark; Landmarks appears in the landmark list and you can move (teleport) to them; You can define the behavior of the teleport: if you will be at a fixed or random distance, looking at the landmark, fixed offset or if it is constrained by a navmesh (when it exists). (optional)
    :param str look_at: The look-at component defines the behavior for an entity to dynamically rotate or face towards another entity or position. Use '#my-camera' to face the user camera, otherwise can take either a vec3 position or a query selector to another entity. (optional)
    :param dict material: The material properties of the object's surface. More properties at <https://aframe.io/docs/1.5.0/components/material.html> A-Frame Material. (optional)
    :param dict material_extras: Define extra material properties, namely texture encoding, whether to render the material's color and render order. The properties set here access directly Three.js material component.  More properties at <https://threejs.org/docs/#api/en/materials/Material> THREE.js Material. (optional)
    :param dict modelUpdate: The GLTF-specific `modelUpdate` attribute is an object with child component names as keys. The top-level keys are the names of the child components to be updated. The values of each are nested `position` and `rotation` attributes to set as new values, respectively. Either `position` or `rotation` can be omitted if unchanged. (optional)
    :param dict multisrc: Define multiple visual sources applied to an object. (optional)
    :param str parent: Parent's object_id. Child objects inherit attributes of their parent, for example scale and translation. (optional)
    :param dict position: 3D object position. (optional)
    :param dict remote_render: Whether or not an object should be remote rendered [Experimental]. (optional)
    :param dict rotation: 3D object rotation in quaternion representation; Right-handed coordinate system. Euler degrees are **deprecated** in wire message format. (optional)
    :param dict scale: 3D object scale. (optional)
    :param bool screenshareable: Whether or not a user can screenshare on an object. Defaults to 'True' (optional)
    :param dict shadow: The shadow component enables shadows for an entity and its children. Adding the shadow component alone is not enough to display shadows in your scene. We must have at least one light with castShadow: true enabled. (optional)
    :param bool show_on_enter_ar: Show object when entering AR. Hidden otherwise. (optional)
    :param bool show_on_enter_vr: Show object when entering VR. Hidden otherwise. (optional)
    :param bool skipCache: Disable retrieving the shared geometry object from the cache. (geometries only: box, circle, cone, ...). Defaults to 'False' (optional)
    :param dict sound: The sound component defines the entity as a source of sound or audio. The sound component can be positional and is thus affected by the component's position. More properties at <https://aframe.io/docs/1.5.0/components/sound.html> A-Frame Sound. (optional)
    :param dict spe_particles: GPU based particle systems in A-Frame. More properties at <https://github.com/harlyq/aframe-spe-particles-component> A-Frame SPE Particles component. (optional)
    :param dict static_body: A fixed-position or animated object. Other objects may collide with static bodies, but static bodies themselves are unaffected by gravity and collisions. More properties at <https://github.com/c-frame/aframe-physics-system/blob/master/CannonDriver.md> A-Frame Physics System. (optional)
    :param str submodel_parent: When this object is parented to a hierarchical model, attaches to a named subcomponent of that model instead of the root position Defaults to '' (optional)
    :param dict textinput: Opens an HTML prompt when clicked. Sends text input as an event on MQTT. Requires click-listener. (optional)
    :param str url: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    :param dict video_control: Adds a video to an entity and controls its playback. (optional)
    :param bool visible: Whether object is visible. Property is inherited. Defaults to 'True' (optional)
    """
    def __init__(self, **kwargs):
        data = {}
        data = Data.update_data(data, kwargs)
        super().__init__(**data)

    @classmethod
    def update_data(cls, data, new_data):
        new_data = new_data.get("data", new_data)
        dash_words = []
        for k, v in new_data.items():

            # when value is None, do not parse it
            if v is None:
                data[k] = v
                continue

            # If the value is another Arena Object besides parent, this is not valid.
            # Crude check that avoids circular import of Arena Object.
            if v.__class__.__bases__[0].__name__ == "Object" and k != 'parent':
                raise ValueError(f"Invalid Arena Object as attribute {k}: {v.object_id}")

            # allow user to input tuples, lists, dicts, etc for specific Attributes.
            # everything gets converted to corresponding attribute
            if (k == "position" or k == "start" or k == "end") and not isinstance(v, Position):
                if isinstance(v, (list, tuple)):
                    data[k] = Position(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Position(**v)
                elif isinstance(v, str):
                    data[k] = Position(v)
                else:
                    data[k] = v

            elif k == "rotation" and not isinstance(v, Rotation):
                if isinstance(v, (list, tuple)):
                    if len(v) == 3:
                        data[k] = Rotation(*v[:3], None)
                    else:
                        data[k] = Rotation(*v[:4])
                elif isinstance(v, dict):
                    data[k] = Rotation(**v)
                elif isinstance(v, str):
                    data[k] = Rotation(v)
                else:
                    data[k] = v

            elif k == "scale" and not isinstance(v, Scale):
                if isinstance(v, (list, tuple)):
                    data[k] = Scale(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Scale(**v)
                elif isinstance(v, str):
                    data[k] = Scale(v)
                else:
                    data[k] = v

            elif k == "color" and not isinstance(v, Color):
                if isinstance(v, (list, tuple)):
                    color = Color(*v[:3])
                elif isinstance(v, dict):
                    color = Color(**v)
                elif isinstance(v, str):
                    color = Color(v)
                else:
                    color = v
                data[k] = color

            elif k == "material":
                if "color" in v:
                    color = v["color"]
                    if isinstance(color, (list, tuple)):
                        color = Color(*color[:3])
                    elif isinstance(color, dict):
                        color = Color(**color)
                    elif isinstance(color, str):
                        color = Color(color)
                    else:
                        color = v["color"]
                    v["color"] = color
                if isinstance(v, dict):
                    data[k] = Material(**v)
                else:
                    data[k] = v

            # Translate and handle underscores from any other keys.
            # Must be done last since KEYWORD_ATTRIBUTE_TRANSLATION contains all attributes,
            # which may interfere with special casing above from, say "rotation" for example.
            elif k in KEYWORD_ATTRIBUTE_TRANSLATION:
                if "-" in k:
                    dash_words += [k]
                    k = KEYWORD_ATTRIBUTE_TRANSLATION[k]
                if isinstance(v, dict):
                    data[k] = ATTRIBUTE_CLASS_TRANSLATION[k](**v)
                else:
                    data[k] = v

            elif isinstance(v, Attribute):
                data[k] = v

            else:
                try:
                    # unknown attribute
                    data[k] = Attribute(**v)
                except:
                    data[k] = v

            # make False into None
            if (isinstance(v, bool) and v == False) or v is None:
                data[k] = None

        # delete elements with keys that have dashes
        for w in dash_words:
            if w in data:
                del data[w]

        return data



from .arena_object import Object

class Cone(Object):
    """
    Cone object class to manage its properties in the ARENA: Cone Geometry.

    :param float height: height Defaults to '1' (optional)
    :param bool openEnded: open ended Defaults to 'False' (optional)
    :param float radiusBottom: radius bottom Defaults to '1' (optional)
    :param float radiusTop: radius top Defaults to '0.01' (optional)
    :param int segmentsHeight: segments height Defaults to '18' (optional)
    :param int segmentsRadial: segments radial Defaults to '36' (optional)
    :param float thetaLength: theta length Defaults to '360' (optional)
    :param float thetaStart: theta start Defaults to '0' (optional)
    """
    object_type = "cone"

    def __init__(self, **kwargs):
        super().__init__(object_type=Cone.object_type, **kwargs)



from .arena_object import Object

class Dodecahedron(Object):
    """
    Dodecahedron object class to manage its properties in the ARENA: Dodecahedron Geometry.

    :param int detail: detail Defaults to '0' (optional)
    :param float radius: radius Defaults to '1' (optional)
    """
    object_type = "dodecahedron"

    def __init__(self, **kwargs):
        super().__init__(object_type=Dodecahedron.object_type, **kwargs)



from .animation import Animation

class AnimationMixer(Animation):
    """
    AnimationMixer attribute class to manage its properties in the ARENA: A list of available animations can usually be found by inspecting the model file or its documentation. All animations will play by default. To play only a specific set of animations, use wildcards: animation-mixer='clip: run_*'. More properties at <https://github.com/n5ro/aframe-extras/tree/master/src/loaders#animation> A-Frame Extras Animation.
    Usage: `animation_mixer=AnimationMixer(...)`

    :param bool clampWhenFinished: If true, halts the animation at the last frame. Defaults to 'False' (optional)
    :param str clip: Name of the animation clip(s) to play. Accepts wildcards. Defaults to '*' (optional)
    :param float crossFadeDuration: Duration of cross-fades between clips, in seconds. Defaults to '0' (optional)
    :param float duration: Duration of the animation, in seconds (0 = auto). Defaults to '0' (optional)
    :param str loop: In repeat and pingpong modes, the clip plays once plus the specified number of repetitions. For pingpong, every second clip plays in reverse. Allows [once, repeat, pingpong] Defaults to 'repeat' (optional)
    :param str repetitions: Number of times to play the clip, in addition to the first play (empty string = Infinity). Repetitions are ignored for loop: once. Defaults to '' (optional)
    :param float startAt: Sets the start of an animation to a specific time (in milliseconds). This is useful when you need to jump to an exact time in an animation. The input parameter will be scaled by the mixer's timeScale. Defaults to '0' (optional)
    :param float timeScale: Scaling factor for playback speed. A value of 0 causes the animation to pause. Negative values cause the animation to play backwards. Defaults to '1' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class Blip(Attribute):
    """
    Blip attribute class to manage its properties in the ARENA: When the object is created or deleted, it will animate in/out of the scene instead of appearing/disappearing instantly. Must have a geometric mesh.
    Usage: `blip=Blip(...)`

    :param bool applyDescendants: Apply blipout effect to include all descendents. Does not work for blipin. Defaults to 'False' (optional)
    :param bool blipin: Animate in on create, set false to disable. Defaults to 'True' (optional)
    :param bool blipout: Animate out on delete, set false to disable. Defaults to 'True' (optional)
    :param float duration: Animation duration in milliseconds. Defaults to '750' (optional)
    :param str geometry: Geometry of the blipout plane. Allows [rect, disk, ring] Defaults to 'rect' (optional)
    :param str planes: Which which clipping planes to use for effect. A top plane clips above it, bottom clips below it. Allows [both, top, bottom] Defaults to 'both' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .arena_object import Object

class Entity(Object):
    """
    Entity object class to manage its properties in the ARENA: Entities are the base of all objects in the scene. Entities are containers into which components can be attached.

    :param dict geometry: geometry (optional)
    :param dict panel: panel (optional)
    """
    object_type = "entity"

    def __init__(self, **kwargs):
        super().__init__(object_type=Entity.object_type, **kwargs)



import math
from collections.abc import Iterable, Mapping

from ..utils import Utils
from .attribute import Attribute


class Rotation(Attribute):
    """
    Rotation attribute class to manage its properties in the ARENA: 3D object rotation in quaternion representation; Right-handed coordinate system. Euler degrees are **deprecated** in wire message format.
    Usage: `rotation=Rotation(...)` or `rotation=Rotation(x,y,z,w)` or `rotation=Rotation(x,y,z)` or `rotation=(x,y,z,w)` or `rotation=(x,y,z)`

    :param float|Iterable|Mapping w: w Defaults to '1' (optional)
    :param float x: x Defaults to '0' (optional)
    :param float y: y Defaults to '0' (optional)
    :param float z: z Defaults to '0' (optional)
    """

    def __init__(self, x=None, y=None, z=None, w=None):
        if isinstance(x, Mapping):
            if "w" in x:
                w = x["w"]
            if "x" in x and "y" in x and "z" in x:
                x, y, z = x["x"], x["y"], x["z"]
            else:
                raise ValueError("Rotation takes x,y,z,(w); a 3-4 element array or list; or a dict with {x,y,z,(w)}")
        elif isinstance(x, str):
            if y is None:
                # Assume that only euler angles are passed as space-delimited string
                x, y, z = map(int, x.split())
            else:
                raise ValueError("Rotation takes x,y,z,(w); a 3-4 element array or list; or a dict with {x,y,z,(w)}")
        elif isinstance(x, Iterable):
            if y is None and 3 <= len(x) <= 4:
                if len(x) == 3:
                    x, y, z = x
                else:
                    x, y, z, w = x
            else:
                raise ValueError("Rotation takes x,y,z,(w); a 3-4 element array or list; or a dict with {x,y,z,(w)}")
        x = x or 0
        y = y or 0
        z = z or 0
        if w is not None:  # quaternion
            self.check_quaternion(x, y, z, w)
            super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z), w=Utils.agran(w))
        else:  # euler
            self.check_euler(x, y, z)
            super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z), w=None)

    def __repr__(self):
        if self.is_quaternion:
            return str(vars(self))
        else:
            rot_dict = vars(self).copy()
            del rot_dict["w"]
            return str(rot_dict)

    @property
    def is_quaternion(self):
        return self.w is not None

    @property
    def euler(self):
        if self.is_quaternion:  # quaternions
            euler = Rotation.q2e((self.x, self.y, self.z, self.w))
            return Rotation(*euler)
        else:  # euler
            return self

    @property
    def quaternion(self):
        if self.is_quaternion:  # quaternions
            return self
        else:  # euler
            quat = Rotation.e2q((self.x, self.y, self.z))
            return Rotation(*quat)

    @property
    def array(self):
        q = self.quaternion
        return [q.x, q.y, q.z, q.w]

    @array.setter
    def array(self, value):
        if not isinstance(value, Iterable) or len(value) not in [3, 4] or isinstance(value, Mapping):
            raise ValueError("Rotation array takes a 3-4 element array or list")
        if len(value) == 3:
            self.check_euler(value, value, value)
            self.x, self.y, self.z = value
            self.w = None
        else:
            self.check_quaternion(value, value, value, value)
            self.x, self.y, self.z, self.w = value

    def check_euler(self, x, y, z):
        values = [x, y, x]
        if any(e < -360 for e in values) or any(e > 360 for e in values):
            raise ValueError(f"Rotation euler values out of range: {x}, {y}, {z}")

    def check_quaternion(self, x, y, z, w):
        values = [x, y, x, w]
        if any(e < -1 for e in values) or any(e > 1 for e in values):
            raise ValueError(f"Rotation quaternion values out of range: {x}, {y}, {z}, {w}")

    @classmethod
    def q2e(cls, q):
        """quaternions to euler (degrees)"""
        x, y, z, w = q

        # Roll (x-axis rotation)
        sinr_cosp = 2.0 * (w * x + y * z)
        cosr_cosp = 1.0 - 2.0 * (x * x + y * y)
        roll = math.atan2(sinr_cosp, cosr_cosp)

        # Pitch (y-axis rotation)
        sinp = 2.0 * (w * y - z * x)
        if abs(sinp) >= 1:
            pitch = math.copysign(math.pi / 2, sinp)  # Use 90 degrees if out of range
        else:
            pitch = math.asin(sinp)

        # Yaw (z-axis rotation)
        siny_cosp = 2.0 * (w * z + x * y)
        cosy_cosp = 1.0 - 2.0 * (y * y + z * z)
        yaw = math.atan2(siny_cosp, cosy_cosp)

        return (math.degrees(roll), math.degrees(pitch), math.degrees(yaw))

    @classmethod
    def e2q(cls, e):
        """euler (degrees) to quaternions"""
        roll, pitch, yaw = e

        roll_rad = math.radians(roll)
        pitch_rad = math.radians(pitch)
        yaw_rad = math.radians(yaw)

        cy = math.cos(yaw_rad * 0.5)
        sy = math.sin(yaw_rad * 0.5)
        cp = math.cos(pitch_rad * 0.5)
        sp = math.sin(pitch_rad * 0.5)
        cr = math.cos(roll_rad * 0.5)
        sr = math.sin(roll_rad * 0.5)

        w = cy * cp * cr + sy * sp * sr
        x = cy * cp * sr - sy * sp * cr
        y = sy * cp * sr + cy * sp * cr
        z = sy * cp * cr - cy * sp * sr

        return (x, y, z, w)



import math
from collections.abc import Iterable, Mapping

from ..utils import Utils
from .attribute import Attribute


class Position(Attribute):
    """
    Position attribute class to manage its properties in the ARENA: 3D object position.
    Usage: `position=Position(...)` or `position=Position(x,y,z)` or `position=(x,y,z)`

    :param float|Iterable|Mapping x: x Defaults to '0' (optional)
    :param float y: y Defaults to '0' (optional)
    :param float z: z Defaults to '0' (optional)
    """

    def __init__(self, x=None, y=None, z=None):
        if isinstance(x, Mapping):
            if "x" in x and "y" in x and "z" in x:
                x, y, z = x["x"], x["y"], x["z"]
            else:
                raise ValueError("Position takes x,y,z; a 3-element array or list; or a dict with {x,y,z}")
        elif isinstance(x, str):
            if y is None:
                x, y, z = map(int, x.split())
            else:
                raise ValueError("Position takes x,y,z; a 3-element array or list; or a dict with {x,y,z}")
        elif isinstance(x, Iterable):
            if y is None and len(x) == 3:
                x, y, z = x
            else:
                raise ValueError("Position takes x,y,z; a 3-element array or list; or a dict with {x,y,z}")

        x = x or 0
        y = y or 0
        z = z or 0
        super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z))

    @property
    def array(self):
        return [self.x, self.y, self.z]

    @array.setter
    def array(self, value):
        if not isinstance(value, Iterable) or len(value) != 3 or isinstance(value, Mapping):
            raise ValueError("Position array takes a 3-element array or list")
        self.x = value[0]
        self.y = value[1]
        self.z = value[2]

    def to_str(self):
        return Utils.tuple_to_string((self.x, self.y, self.z))

    def distance_to(self, pos):
        return math.sqrt((self.x - pos.x) ** 2 + (self.y - pos.y) ** 2 + (self.z - pos.z) ** 2)



from .attribute import Attribute


class Armarker(Attribute):
    """
    Armarker attribute class to manage its properties in the ARENA: A location marker (such as an AprilTag, a lightAnchor, or an UWB tag), used to anchor scenes, or scene objects, in the real world.
    Usage: `armarker=Armarker(...)`

    :param bool buildable: Whether tag has 'dynamic' toggled on click. Used to position a tag, then lock into position. Defaults to 'False' (optional)
    :param bool dynamic: Dynamic tag, not used for localization. E.g., to move object to which this ARMarker component is attached to. Requires permissions to update the scene (if dynamic=true). Defaults to 'False' (optional)
    :param float ele: Tag elevation in meters. Defaults to '0' (optional)
    :param float lat: Tag latitude. Defaults to '0' (optional)
    :param float long: Tag longitude. Defaults to '0' (optional)
    :param str markerid: The marker id (e.g. for AprilTag 36h11 family, an integer in the range [0, 586]). Defaults to '0' (optional)
    :param str markertype: The marker type, technology-based. Allows [apriltag_36h11, lightanchor, uwb, vive, optitrack] Defaults to 'apriltag_36h11' (optional)
    :param bool publish: Publish detections. Send detections to external agents (e.g. external builder script that places new markers in the scene). If dynamic=true and publish=true, object position is not updated (left up to external agent). Defaults to 'False' (optional)
    :param float size: Tag size in millimeters. Defaults to '150' (optional)
    :param str url: URL associated with the tag. Defaults to '' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .arena_object import Object

class ArenauiButtonPanel(Object):
    """
    ArenauiButtonPanel object class to manage its properties in the ARENA: ARENAUI element which displays a vertical or horizontal panel of buttons.

    :param list[dict] buttons: Buttons Defaults to '[{'name': 'Option 1'}, {'name': 'Option 2'}]' (optional)
    :param str font: Font to use for button text. Allows [Roboto, Roboto-Mono] Defaults to 'Roboto' (optional)
    :param str materialSides: Which sides display the rendered UI material Allows [both, front] Defaults to 'both' (optional)
    :param str theme: Color Theme Allows [light, dark] Defaults to 'light' (optional)
    :param str title: Title to display above buttons (optional). Defaults to '' (optional)
    :param bool vertical: Vertical button layout Defaults to 'False' (optional)
    """
    object_type = "arenaui-button-panel"

    def __init__(self, **kwargs):
        super().__init__(object_type=ButtonPanel.object_type, **kwargs)

    def json_preprocess(self, **kwargs):
        # kwargs are for additional param to add to json, like "action":"create"
        skipped_keys = [
            "evt_handler",
            "update_handler",
            "animations",
            "delayed_prop_tasks",
        ]
        json_payload = {k: v for k, v in vars(
            self).items() if k not in skipped_keys}
        json_payload["data"]["buttons"] = [
            vars(button) if hasattr(button, "__dict__") else button
            for button in json_payload["data"]["buttons"]
        ]
        json_payload.update(kwargs)
        return json_payload


class ButtonPanel(ArenauiButtonPanel):
    """
    Alternate name for ArenauiButtonPanel.
    """


class Button:
    """
    Buttons of a ButtonPanel in the ARENA UI.

    :param str name: Button Name
    :param str img: Image URl for image buttons (optional)
    :param float size: Size of image button (optional)
    :param float height: Height of image button, overrides size (optional)
    :param float: Width of image button, overrides size (optional)
    :param float borderRadius: Border radius of image button (optional)
    """

    def __init__(self, name="Button", **kwargs):
        self.name = name
        self.__dict__.update(kwargs)



from .attribute import Attribute


class Textinput(Attribute):
    """
    Textinput attribute class to manage its properties in the ARENA: Opens an HTML prompt when clicked. Sends text input as an event on MQTT. Requires click-listener.
    Usage: `textinput=Textinput(...)`

    :param str label: Text prompt label Defaults to 'Input text below (max is 140 characters).' (optional)
    :param str on: A case-sensitive string representing the event type to listen for. See <https://developer.mozilla.org/en-US/docs/Web/Events> Web Events Allows [mousedown, mouseup, mouseenter, mouseleave, triggerdown, triggerup, gripdown, gripup, menudown, menuup, systemdown, systemup, trackpaddown, trackpadup] Defaults to 'mousedown' (optional)
    :param str placeholder: Text input place holder. Defaults to 'Type here' (optional)
    :param str title: The prompt title. Defaults to 'Text Input' (optional)
    """

    def __init__(self,
                 title="Text Input",
                 label="Input text below (max is 140 characters)",
                 placeholder="Type here",
                 on="mousedown",
                 **kwargs):
        super().__init__(title=title[:140], label=label[:140],
                         placeholder=placeholder[:140], on=on, **kwargs)


class TextInput(Textinput):
    """
    Alternate name for Textinput.
    Usage: `textinput=TextInput(...)`
    """



from deprecated import deprecated

from .arena_object import Object


class Text(Object):
    """
    Text object class to manage its properties in the ARENA: Display text. More properties at <https://aframe.io/docs/1.5.0/components/text.html> A-Frame Text.

    :param str align: Multi-line text alignment. Allows [left, center, right] Defaults to 'left' (optional)
    :param float alphaTest: Discard text pixels if alpha is less than this value. Defaults to '0.5' (optional)
    :param str anchor: Horizontal positioning. Allows [left, right, center, align] Defaults to 'center' (optional)
    :param str baseline: Vertical positioning. Allows [top, center, bottom] Defaults to 'center' (optional)
    :param str color: Text color. Defaults to '#000000' (optional)
    :param str font: Font to render text, either the name of one of A-Frame's stock fonts or a URL to a font file. Allows [aileronsemibold, dejavu, exo2bold, exo2semibold, kelsonsans, monoid, mozillavr, roboto, sourcecodepro] Defaults to 'roboto' (optional)
    :param str fontImage: Font image texture path to render text. Defaults to the font's name with extension replaced to .png. Don't need to specify if using a stock font. (derived from font name). (optional)
    :param float height: Height of text block. (derived from text size). (optional)
    :param float letterSpacing: Letter spacing in pixels. Defaults to '0' (optional)
    :param float lineHeight: Line height in pixels. (derived from font file). (optional)
    :param float opacity: Opacity, on a scale from 0 to 1, where 0 means fully transparent and 1 means fully opaque. Defaults to '1' (optional)
    :param str shader: Shader used to render text. Allows [portal, flat, standard, sdf, msdf, ios10hls, skyshader, gradientshader] Defaults to 'sdf' (optional)
    :param str side: Side to render. Allows [front, back, double] Defaults to 'double' (optional)
    :param float tabSize: Tab size in spaces. Defaults to '4' (optional)
    :param str text: ***DEPRECATED**: data.text is **deprecated** for object_type: text, use data.value instead.* (deprecated)
    :param bool transparent: Whether text is transparent. Defaults to 'True' (optional)
    :param str value: The actual content of the text. Line breaks and tabs are supported with `\\n` and `\\t`. (optional)
    :param str whiteSpace: How whitespace should be handled. Allows [normal, pre, nowrap] Defaults to 'normal' (optional)
    :param float width: Width in meters. (derived from geometry if exists). Defaults to '5' (optional)
    :param float wrapCount: Number of characters before wrapping text (more or less). Defaults to '40' (optional)
    :param float wrapPixels: Number of pixels before wrapping text. (derived from wrapCount). (optional)
    :param float xOffset: X-offset to apply to add padding. Defaults to '0' (optional)
    :param float zOffset: Z-offset to apply to avoid Z-fighting if using with a geometry as a background. Defaults to '0.001' (optional)
    """
    object_type = "text"

    def __init__(self, **kwargs):
        # NOTE: Don't require parameter 'text' or 'value' we won't know which one users
        #       used downstream into persist
        super().__init__(object_type=Text.object_type, **kwargs)

    @property
    @deprecated("DEPRECATED:data.text is deprecated for object_type: text, use data.value instead.")
    def text(self):
        return None

    @text.setter
    @deprecated("DEPRECATED:data.text is deprecated for object_type: text, use data.value instead.")
    def text(self, value):
        return



from .attribute import Attribute


class ModelUpdate(Attribute):
    """
    ModelUpdate attribute class to manage its properties in the ARENA: The GLTF-specific `modelUpdate` attribute is an object with child component names as keys. The top-level keys are the names of the child components to be updated. The values of each are nested `position` and `rotation` attributes to set as new values, respectively. Either `position` or `rotation` can be omitted if unchanged.
    Usage: `modelUpdate=ModelUpdate(...)`

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import re

import webcolors

from .attribute import Attribute


class Color(Attribute):
    """
    Color Attribute.
    Usage: `color=Color(red,green,blue)` or `color=(red,green,blue)`
    """

    def __init__(self, red=0, green=0, blue=0):
        if isinstance(red, str):
            # hex to tuple to Color
            color = red.lstrip("#")
            hexcolor = re.search(r"^(?:[0-9a-fA-F]{3}){1,2}$", color)
            wcrgb = None
            if not hexcolor:
                # try appending leading zeros until it works
                for i in range(8):
                    try:
                        wcrgb = webcolors.name_to_rgb("#" + i * "0" + color)
                    except ValueError:
                        break
                if wcrgb is not None:
                    c = (wcrgb.red, wcrgb.green, wcrgb.blue)
                else:
                    c = (128, 128, 128)
            else:
                c = tuple(int(color[c:c + 2], 16) for c in (0, 2, 4))
            self.red, self.green, self.blue = c

        if isinstance(red, (float, int)):
            self.red = int(red)
        if isinstance(blue, (float, int)):
            self.blue = int(blue)
        if isinstance(green, (float, int)):
            self.green = int(green)

        super().__init__(red=self.red, green=self.green, blue=self.blue)

    @property
    def hex(self):
        if isinstance(self.red, str):
            return self.red
        else:
            return f"#{''.join(f'{i:02x}' for i in [self.red, self.green, self.blue])}"



from .attribute import Attribute


class GltfModelLod(Attribute):
    """
    GltfModelLod attribute class to manage its properties in the ARENA: Simple switch between the default gltf-model and a detailed one when a user camera is within specified distance
    Usage: `gltf_model_lod=GltfModelLod(...)`

    :param float detailedDistance: At what distance to switch between the models. Defaults to '10' (optional)
    :param str detailedUrl: Alternative 'detailed' gltf model to load by URL. Defaults to '' (optional)
    :param bool retainCache: Whether to skip freeing the detailed model from browser cache (default false). Defaults to 'False' (optional)
    :param float updateRate: How often user camera is checked for LOD (default 333ms). Defaults to '333' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .arena_object import Object


class Box(Object):
    """
    Box object class to manage its properties in the ARENA: Box Geometry.

    :param float depth: depth Defaults to '1' (optional)
    :param float height: height Defaults to '1' (optional)
    :param int segmentsDepth: segments depth Defaults to '1' (optional)
    :param int segmentsHeight: segments height Defaults to '1' (optional)
    :param int segmentsWidth: segments width Defaults to '1' (optional)
    :param float width: width Defaults to '1' (optional)
    """
    object_type = "box"

    def __init__(self, **kwargs):
        super().__init__(object_type=Box.object_type, **kwargs)



import asyncio
import json
import os
import random
import socket
import ssl
import sys
from datetime import datetime

import paho.mqtt.client as mqtt

from .auth import ArenaAuth
from .env import ARENA_PASSWORD, ARENA_USERNAME, MQTTH, NAMESPACE, REALM, _get_env
from .event_loop import *
from .topics import PUBLISH_TOPICS, SUBSCRIBE_TOPICS, TOPIC_TYPES


class ArenaMQTT(object):
    """
    Wrapper around Paho MQTT client and EventLoop.
    """

    scene = None
    device = None
    auth = None

    def __init__(
                self,
                web_host = "arenaxr.org",
                realm = "realm",
                network_latency_interval = 10000,  # run network latency update every 10s
                on_msg_callback = None,
                end_program_callback = None,
                video = False,
                debug = False,
                headless = False,
                environment = False,
                **kwargs
            ):
        if os.environ.get(MQTTH):
            self.web_host = _get_env(MQTTH)
            print(f"Using Host from 'MQTTH' env variable: {self.web_host}")
        elif "host" in kwargs and kwargs["host"]:
            self.web_host = kwargs["host"]
            print(f"Using Host from 'host' input parameter: {self.web_host}")
        else:
            # Use default "web_host", helps avoid and web vs mqtt host and other user setup confusion
            self.web_host = web_host

        if os.environ.get(REALM):
            self.realm = _get_env(REALM)
            print(f"Using Realm from 'REALM' env variable: {self.realm}")
        elif "realm" in kwargs and kwargs["realm"]:
            self.realm = kwargs["realm"]
            print(f"Using Realm from 'realm' input parameter: {self.realm}")
        else:
            # Use default "realm" until multiple realms exist, avoids user confusion.
            self.realm = realm

        self.debug = debug

        print("=====")
        # do user auth
        self.username = None
        token = None
        self.remote_auth_token = {}  # provide reference for downloaded token
        self.auth = ArenaAuth()
        if os.environ.get(ARENA_USERNAME) and os.environ.get(ARENA_PASSWORD):
            # auth 1st: use passed in env var
            self.username = _get_env(ARENA_USERNAME)
            token = _get_env(ARENA_PASSWORD)
            self.auth.store_environment_auth(self.username, token)
        else:
            if self.scene:
                local = self.auth.check_local_auth()
            elif self.device:
                local = self.auth.authenticate_device(self.web_host)
            if local and "username" in local and "token" in local:
                # auth 2nd: use locally saved token
                self.username = local["username"]
                token = local["token"]
            else:
                if self.scene:
                    # auth 3rd: use the user account online
                    self.username = self.auth.authenticate_user(self.web_host, headless)

        if os.environ.get(NAMESPACE):
            self.namespace = _get_env(NAMESPACE)
        elif "namespace" not in kwargs or ("namespace" in kwargs and kwargs["namespace"] is None):
            self.namespace = self.username
        else:
            self.namespace = kwargs["namespace"]

        if self.scene:
            self.namespaced_target = f"{self.namespace}/{self.scene}"
        elif self.device:
            self.namespaced_target = f"{self.namespace}/{self.device}"

        # Always use the the hostname specified by the user, or defaults.
        self.mqttc_id = "pyClient-" + self.generate_client_id()

        # fetch host config
        print("Fetching ARENA configuration...")
        self.config_url = f"https://{self.web_host}/conf/defaults.json"
        self.config_data = json.loads(self.auth.urlopen(self.config_url))

        self.mqtt_host = self.config_data["ARENADefaults"]["mqttHost"]

        if self.scene and (not self.username or not token):
            # do scene auth by user
            data = self.auth.authenticate_scene(
                self.web_host, self.realm, self.namespaced_target, self.username, video, environment
            )
            if "username" in data and "token" in data and "ids" in data:
                self.username = data["username"]
                token = data["token"]
                self.remote_auth_token = data

        # prefer user_id from account, however root tokens use mqtt client_id as a substitute
        if self.remote_auth_token and "ids" in self.remote_auth_token and "userid" in self.remote_auth_token["ids"]:
            self.userid = self.remote_auth_token["ids"]["userid"]
            self.userclient = self.remote_auth_token["ids"]["userclient"]
        else:
            self.userid = self.mqttc_id
            self.userclient = self.mqttc_id

        # set up topic variables
        self.topicParams = {  # Reusable topic param dict
            "realm": self.realm,
            "nameSpace": self.namespace,
            # adding "sceneName" or "deviceName" depending on case below
            "userClient": self.userclient,
            "idTag": self.userid,
        }
        if self.scene:
            self.topicParams["sceneName"] = self.scene
            self.subscribe_topics = {
                'public': SUBSCRIBE_TOPICS.SCENE_PUBLIC.substitute(self.topicParams),
                'private': SUBSCRIBE_TOPICS.SCENE_PRIVATE.substitute(self.topicParams)
            }
            if environment:
                self.subscribe_topics["envhost"] = SUBSCRIBE_TOPICS.SCENE_ENV_PRIVATE.substitute(
                    {**self.topicParams, **{"idTag": "-"}}
                )
        elif self.device:
            self.topicParams["deviceName"] = self.device
            self.subscribe_topics = {
                'public': SUBSCRIBE_TOPICS.DEVICE.substitute(self.topicParams),
            }
        self.latency_topic = self.config_data["ARENADefaults"]["latencyTopic"]  # network graph latency update
        self.ignore_topic = SUBSCRIBE_TOPICS.SCENE_PUBLIC_SELF.substitute(self.topicParams)

        # check for valid permissions to write to all objects topics
        self.can_publish_obj = self.auth.has_publish_rights(
            token,
            PUBLISH_TOPICS.SCENE_OBJECTS.substitute({**self.topicParams, **{"objectId": "anyfoobject"}})
        )

        self.mqttc = mqtt.Client(
            self.mqttc_id, clean_session=True
        )
        self.mqttc.username_pw_set(username=self.username, password=token)
        print("=====")

        # set up callbacks
        self.on_msg_callback = on_msg_callback
        self.end_program_callback = end_program_callback

        self.event_loop = EventLoop(self.disconnect)

        aioh = AsyncioMQTTHelper(self.event_loop, self.mqttc)

        # have all tasks wait until mqtt client is connected before starting
        self.mqtt_connect_evt = asyncio.Event()
        self.mqtt_connect_evt.clear()

        # set paho mqtt callbacks
        self.subscriptions = {}
        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect
        self.mqttc.on_publish = self.on_publish
        self.mqttc.on_subscribe = self.on_subscribe

        # add main message processing + callbacks loop to tasks
        self.run_async(self.process_message)

        # update network latency every network_latency_interval secs
        self.run_forever(self.network_latency_update,
                         interval_ms=network_latency_interval)

        self.msg_queue = asyncio.Queue()

        # connect to mqtt broker
        if "port" in kwargs:
            port = kwargs["port"]
        else:
            port = 8883  # ARENA broker TLS 1.2 connection port
        if self.auth.verify(self.web_host):
            self.mqttc.tls_set()
        else:
            self.mqttc.tls_set_context(ssl._create_unverified_context())
            self.mqttc.tls_insecure_set(True)
        try:
            self.mqttc.connect(self.mqtt_host, port=port, keepalive=60)
        except Exception as err:
            print(f'MQTT connect error to {self.mqtt_host}, port={port}: Result Code={err}')
        self.mqttc.socket().setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)

    def generate_client_id(self):
        """Returns a random 6 digit id."""
        return str(random.randrange(100000, 999999))

    def network_latency_update(self):
        """Update client latency in $NETWORK/latency."""
        # publish empty message with QoS of 2 to update latency
        self.mqttc.publish(self.latency_topic, "", qos=2)

    def run_once(self, func=None, **kwargs):
        """Runs a user-defined function on startup."""
        if func is not None:
            w = SingleWorker(self.event_loop, func, self.mqtt_connect_evt, **kwargs)
            self.event_loop.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_once(func):
                self.run_once(func, **kwargs)
                return func
            return _run_once

    def run_after_interval(self, func=None, interval_ms=1000, **kwargs):
        """Runs a user-defined function after a interval_ms milliseconds."""
        if func is not None:
            if interval_ms < 0:
                print("Invalid interval! Defaulting to 1000ms")
                interval_ms = 1000
            w = LazyWorker(self.event_loop, func, self.mqtt_connect_evt, interval_ms, **kwargs)
            self.event_loop.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_after_interval(func):
                self.run_after_interval(func, interval_ms, **kwargs)
                return func
            return _run_after_interval

    def run_async(self, func=None, **kwargs):
        """Runs a user defined aynscio function."""
        if func is not None:
            w = AsyncWorker(self.event_loop, func, self.mqtt_connect_evt, **kwargs)
            self.event_loop.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_async(func):
                self.run_async(func, **kwargs)
                return func
            return _run_async

    def run_forever(self, func=None, interval_ms=1000, **kwargs):
        """Runs a function every interval_ms milliseconds."""
        if func is not None:
            if interval_ms < 0:
                print("Invalid interval! Defaulting to 1000ms")
                interval_ms = 1000
            t = PersistentWorker(self.event_loop, func, self.mqtt_connect_evt, interval_ms, **kwargs)
            self.event_loop.add_task(t)
        else:
            # if there is no func, we are in a decorator
            def _run_forever(func):
                self.run_forever(func, interval_ms, **kwargs)
                return func
            return _run_forever

    def run_tasks(self):
        """Run event loop."""
        print("Connecting to the ARENA...")
        self.event_loop.run()

    def stop_tasks(self):
        """Stop event loop."""
        self.event_loop.stop()

    async def sleep(self, interval_ms):
        """Public function for sleeping in async functions."""
        await asyncio.sleep(interval_ms / 1000)

    def on_connect(self, client, userdata, flags, rc):
        """Paho MQTT client on_connect callback."""
        if rc == 0:
            self.mqtt_connect_evt.set()

            # listen to all messages in scene
            self.do_subscribe(client, self.subscribe_topics['public'], self.on_message)
            if self.subscribe_topics.get('private'):
                self.do_subscribe(client, self.subscribe_topics['private'], self.on_message_private)

            # reset msg rate time
            self.msg_rate_time_start = datetime.now()

            print("Connected!")
            print("=====")

        else:
            print(f"Connection error! Result code={rc}")
            os._exit(1)

    def do_subscribe(self, client, topic, callback):
        result, mid = client.subscribe(topic)
        if result == mqtt.MQTT_ERR_SUCCESS:
            self.subscriptions[mid] = topic
        else:
            print(f"Subscribe ERROR!!! topic={topic} result code={result}")
        client.message_callback_add(topic, callback)

    def on_message(self, client, userdata, msg):
        # ignore own messages
        if mqtt.topic_matches_sub(self.ignore_topic, msg.topic):
            return
        self.msg_queue.put_nowait(msg)

    def on_message_private(self, client, userdata, msg):
        # Private messages are never reflected to self, no check required
        self.msg_queue.put_nowait(msg)

    async def process_message(self):
        raise NotImplementedError("Must override process_message")

    def on_subscribe(self, client, userdata, mid, granted_qos, properties=None):
        if self.debug:
            print(f"[subscribe ack]: topic={self.subscriptions[mid]} granted qos={granted_qos}")
        for qos in granted_qos:
            if qos not in range(0, 2):
                if mid in self.subscriptions:
                    print(f"FAILURE!!! Subscribing to topic {self.subscriptions[mid]}")
                else:
                    print(f"FAILURE!!! Subscribing to topic with message id: {mid}")

    def on_disconnect(self, client, userdata, rc):
        """Paho MQTT client on_disconnect callback."""
        if rc == 0:
            print("Disconnected from the ARENA!")
        else:
            print(f"Disconnected! Result code={rc}.")

        os._exit(rc)

    def disconnect(self):
        """Disconnects Paho MQTT client."""
        if self.end_program_callback:
            self.end_program_callback(self)
        self.mqttc.disconnect()

    def message_callback_add(self, sub, callback):
        """Subscribes to new topic and adds callback."""
        self.do_subscribe(self.mqttc, sub, callback)

    def message_callback_remove(self, sub):
        """Unsubscribes to topic and removes callback."""
        self.mqttc.unsubscribe(sub)
        self.mqttc.message_callback_remove(sub)

    def rcv_queue_len(self):
        """Return receive queue length."""
        self.msg_queue.qsize()

    def pub_queue_len(self):
        """Return publish queue length."""
        return self.mqttc._out_packet

    def client_id(self):
        """Return client id."""
        return self.mqttc_id



from .arena_object import Object

class Capsule(Object):
    """
    Capsule object class to manage its properties in the ARENA: Capsule Geometry.

    :param float length: length Defaults to '1' (optional)
    :param float radius: radius Defaults to '1' (optional)
    :param int segmentsCap: segments capsule Defaults to '18' (optional)
    :param int segmentsRadial: segments radial Defaults to '36' (optional)
    """
    object_type = "capsule"

    def __init__(self, **kwargs):
        super().__init__(object_type=Capsule.object_type, **kwargs)



from .attribute import Attribute

class GotoUrl(Attribute):
    """
    GotoUrl attribute class to manage its properties in the ARENA: Load new URL when object is clicked. Requires click-listener.
    Usage: `goto_url=GotoUrl(...)`

    :param str dest: Where to open the URL. Allows [popup, newtab, sametab] Defaults to 'sametab' (optional)
    :param str on: A case-sensitive string representing the event type to listen for. See <https://developer.mozilla.org/en-US/docs/Web/Events> Web Events. Allows [mousedown, mouseup] Defaults to 'mousedown' (optional)
    :param str url: The destination url, e.g. https://example.com. Defaults to '' (optional)
    """
    def __init__(self, url, dest="static", on="mousedown", **kwargs):
        if dest != "popup" and dest != "newtab" and dest != "sametab":
            dest = "popup"
        super().__init__(url=url, dest=dest, on=on, **kwargs)



from .arena_object import Object

class HandRight(Object):
    """
    HandRight object class to manage its properties in the ARENA: Hand is the (left or right) hand metadata pose and controller type of the user avatar.

    :param str dep: Camera object_id this hand belongs to. Defaults to '' (optional)
    :param str url: Path to user avatar hand model. Defaults to 'static/models/hands/valve_index_left.gltf' (optional)
    """
    object_type = "handRight"

    def __init__(self,**kwargs):
        self.camera = None
        super().__init__(object_type=HandRight.object_type, **kwargs)



from .arena_object import Object

class ThreejsScene(Object):
    """
    ThreejsScene object class to manage its properties in the ARENA: Load a Three.js Scene. Could be THREE.js version-specific; you can see the THREE.js version in the JS console once you open ARENA; using glTF is preferred. Format: <https://threejs.org/docs/#api/en/scenes/Scene> THREE.js Scene. See guidance to store paths under <https://docs.arenaxr.org/content/interface/filestore.html> ARENA File Store, CDN, or DropBox.

    :param str url: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    """
    object_type = "threejs-scene"

    def __init__(self, **kwargs):
        super().__init__(object_type=ThreejsScene.object_type, **kwargs)



import argparse
import asyncio
import json
import os
import re
import sys
import threading
import traceback
import uuid
from datetime import datetime
from inspect import signature
from pathlib import Path

import __main__ as main

from .arena_mqtt import ArenaMQTT
from .attributes import *
from .env import PROGRAM_OBJECT_ID, SCENE, _get_env
from .events import *
from .objects import *
from .topics import PUBLISH_TOPICS, SCENE_MSGTYPES, TOPIC_TOKENS, TOPIC_TYPES
from .utils import (
    ArenaCmdInterpreter,
    ArenaTelemetry,
    ProgramRunInfo,
    QueueStats,
    Utils,
)


class Scene(ArenaMQTT):
    """
    Gives access to an ARENA scene.
    Can create and execute various user-defined functions/tasks.

    :param str host: Hostname of the ARENA webserver (required).
    :param str realm: Reserved topic fork for future use (optional).
    :param str namespace: Username of authenticated user or other namespace (automatic).
    :param str scene: The name of the scene, without namespace (required).
    :param int network_latency_interval: Interval (in ms) to run network graph latency update. Default value is 10000 (10 secs). Ignore this parameter (optional).
    :param func on_msg_callback: Called on all MQTT messages received (optional).
    :param func new_obj_callback: Called on object 'create' MQTT messages received (optional).
    :param func user_join_callback: Called on user id 'new' MQTT messages received (optional).
    :param func user_left_callback: Called on user id 'leave' MQTT messages received (optional).
    :param func delete_obj_callback: Called on object 'delete' MQTT messages received (optional).
    :param func end_program_callback: Called on MQTT disconnect (optional).
    :param bool video: If true, request permissions for video conference (optional).
    :param bool debug: If true, print a log of all publish messages from this client (optional).
    :param bool cli_args: If true, require CLI standardized parameters (optional).
    :param bool headless: If true, force limited input device auth flow (optional).
    """

    def __init__(
        self,
        host="arenaxr.org",
        realm="realm",
        network_latency_interval=10000,  # run network latency update every 10s
        on_msg_callback=None,
        new_obj_callback=None,
        user_join_callback=None,
        user_left_callback=None,
        delete_obj_callback=None,
        end_program_callback=None,
        video=False,
        debug=False,
        cli_args=False,
        headless=False,
        **kwargs,
    ):

        # init telemetry
        self.telemetry = ArenaTelemetry()

        # setup event to let others wait on connection
        self.connected_evt = threading.Event()

        # start the command interpreter (if enabled by env variable)
        self.cmd_interpreter = ArenaCmdInterpreter(
            self,
            show_attrs=("config_data", "scene", "users", "all_objects", "run_info"),
            get_callables=("persisted_objs", "persisted_scene_option", "writable_scenes", "user_list"),
            start_cmd_event=self.connected_evt,
        )

        if cli_args:
            self.args = self.parse_cli(cli_args)
            if self.args["host"]:
                kwargs["host"] = self.args["host"]
            if self.args["namespace"]:
                kwargs["namespace"] = self.args["namespace"]
            if self.args["scene"]:
                kwargs["scene"] = self.args["scene"]
            if self.args["debug"]:
                debug = self.args["debug"]

        if os.environ.get(SCENE):
            self.scene = _get_env(SCENE)
            print(f"Using Scene from 'SCENE' env variable: {self.scene}")
        elif "scene" in kwargs and kwargs["scene"]:
            self.scene = kwargs["scene"]
            print(f"Using Scene from 'scene' input parameter: {self.scene}")
        else:
            self.exit("Scene argument (scene) is unspecified or None, aborting...")

        if re.search("/", self.scene):
            self.exit("Scene cannot include '/', aborting...")

        super().__init__(
            host,
            realm,
            network_latency_interval,
            on_msg_callback,
            end_program_callback,
            video,
            debug,
            headless,
            **kwargs,
        )

        with self.telemetry.start_span(f"init {self.namespace}/{self.scene}") as span:
            # 'init' span will track the remainder of the initialization

            # create a program object to describe this program
            # PROGRAM_OBJECT_ID allows to match the object id of persisted program object
            # when a program object with PROGRAM_OBJECT_ID is loaded from persist, it will replace this one
            self.program = Program(
                object_id=_get_env(PROGRAM_OBJECT_ID, super().client_id()),
                name=f"{self.namespace}/{self.scene}",
                filename=sys.argv[0],
                filetype="PY",
            )

            self.persist_host = self.config_data["ARENADefaults"]["persistHost"]
            self.persist_path = self.config_data["ARENADefaults"]["persistPath"]

            self.persist_url = f"https://{self.persist_host}{self.persist_path}{self.namespaced_target}"

            # set up callbacks
            self.new_obj_callback = new_obj_callback
            self.delete_obj_callback = delete_obj_callback
            self.user_join_callback = user_join_callback
            self.user_left_callback = user_left_callback

            # objects that exist in the scene, but this scene instance does not have a reference to
            self.unspecified_object_ids = set()
            self.users = {}  # dict of all users

            # setup program run info to collect stats
            self.run_info = ProgramRunInfo(
                self.event_loop,
                queue_len_callable=lambda: self.get_rcv_pub_queue_len(),
                update_callback=self.run_info_update,
                web_host=self.web_host,
                namespace=self.namespace,
                scene=self.scene,
                realm=self.realm,
            )

            # Always use the the hostname specified by the user, or defaults.
            print(f"Loading: https://{self.web_host}/{self.namespace}/{self.scene}, realm={self.realm}")

            span.add_event(f"Loading: https://{self.web_host}/{self.namespace}/{self.scene}, realm={self.realm}")

    def parse_cli(self, cli_args=False):
        """
        Reusable command-line options to give apps flexible options to avoid hard-coding locations.
        """
        parser = argparse.ArgumentParser(
            description=(f"{Path(main.__file__).name} (arena-py) Application CLI"),
            epilog="Additional user-defined args are possible, see docs at https://docs.arenaxr.org/content/python/scenes for usage.",
        )
        parser.add_argument("-mh", "--host", type=str, help="ARENA webserver main host to connect to")
        parser.add_argument("-n", "--namespace", type=str, help="Namespace of scene")
        parser.add_argument("-s", "--scene", type=str, help="Scene to publish and listen to")
        parser.add_argument("-d", "--device", type=str, help="Device to publish and listen to")
        parser.add_argument("-p", "--position", nargs=3, type=float, default=(0, 0, 0), help="App position as 'x y z'")
        parser.add_argument("-r", "--rotation", nargs=3, type=float, default=(0, 0, 0), help="App rotation as 'x y z'")
        parser.add_argument("-c", "--scale", nargs=3, type=float, default=(0, 0, 0), help="App scale as 'x y z'")
        parser.add_argument("-D", "--debug", action="store_true", help="Debug mode.", default=False)

        # add known help descriptions
        if isinstance(cli_args, dict):
            for k in cli_args:
                parser.add_argument(f"-{k}", f"--{k}", help=cli_args[k], required=False)

        # add unknown arguments for users to pull as strings
        parsed, unknown = parser.parse_known_args()
        for arg in unknown:
            if arg.startswith(("-", "--")):
                parser.add_argument(arg.split("=")[0])

        args = parser.parse_args()
        argdict = vars(args)
        argdict["position"] = tuple(args.position)
        argdict["rotation"] = tuple(args.rotation)
        argdict["scale"] = tuple(args.scale)
        return argdict

    def exit(self, arg=None):
        """Custom exit to push errors to telemetry."""
        error_msg = None
        if arg != None and arg != 0:
            error_msg = f"Exiting with sys.exit('{arg}')"
        self.telemetry.exit(error_msg)
        os._exit(0 if error_msg == None else 1)

    def on_connect(self, client, userdata, flags, rc):
        super().on_connect(client, userdata, flags, rc)
        if rc == 0:
            # set event
            self.connected_evt.set()

            # create arena-py Objects from persist server
            # no need to return anything here
            self.get_persisted_objs()

    def on_message(self, client, userdata, msg):
        super().on_message(client, userdata, msg)
        self.run_info.msg_rcv()

    def on_message_private(self, client, userdata, msg):
        super().on_message_private(client, userdata, msg)
        self.run_info.msg_rcv()

    def on_publish(self, client, userdata, mid):
        self.run_info.msg_publish()

    async def process_message(self):
        while True:
            try:
                msg = await self.msg_queue.get()
            except RuntimeError as e:
                self.telemetry.add_event(f"Ignoring error: {e}")
                return

            # extract payload
            try:
                payload_str = msg.payload.decode("utf-8", "ignore")
                payload = json.loads(payload_str)
            except Exception as e:
                self.telemetry.add_event(f"Malformed payload: {payload_str}. {e}.")
                return

            if self.debug:
                # log messages received for debugging
                self.telemetry.add_event(f"[received] {msg.topic} {payload}")

            payload["topic"] = msg.topic  # add topic for downstream callbacks/handlers
            object_id = payload.get("object_id", None)
            action = payload.get("action", None)
            topic_split = msg.topic.split("/")

            # Validate scene messages object_id vs topic uuid token
            scene_msgtype = None
            if topic_split[TOPIC_TOKENS.TYPE] == TOPIC_TYPES.SCENE:
                topic_uuid = topic_split[TOPIC_TOKENS.UUID]
                if object_id != topic_uuid:
                    self.telemetry.set_error(f"Message object_id {object_id} does not match topic {topic_uuid}.")
                    return
                scene_msgtype = topic_split[TOPIC_TOKENS.SCENE_MSGTYPE]
                # Object updates only in these scene msg types
                if scene_msgtype not in [SCENE_MSGTYPES.PRESENCE, SCENE_MSGTYPES.USER, SCENE_MSGTYPES.OBJECTS, SCENE_MSGTYPES.PROGRAM]:
                    return

            with self.telemetry.start_process_msg_span(object_id, action) as span:
                try:
                    # update object attributes
                    data = payload.get("data", {})
                    object_type = data.get("object_type", None)

                    event = None

                    # create/get object from object_id
                    if object_id in self.all_objects:
                        obj = self.all_objects[object_id]
                    else:
                        ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                        obj = ObjClass(**payload)

                    # react to action accordingly
                    if action:
                        if action == "clientEvent":
                            event = Event(**payload)
                            # get object from target
                            if "target" in data and data["target"] in self.all_objects:
                                obj = self.all_objects[data["target"]]
                                if obj.evt_handler:
                                    self.callback_wrapper(obj.evt_handler, event, payload)
                                    continue
                            span.add_event("Client event: {event}")
                        elif action == "leave":
                            # special user presence message, new way to remove user
                            if object_id in self.users:
                                self.delete_user_objects(object_id)
                                if self.user_left_callback:
                                    self.callback_wrapper(self.user_left_callback, self.users[object_id], payload)
                                del self.users[object_id]
                            span.add_event("User left.")
                        elif action == "delete":
                            if Camera.object_type in object_id:  # object is a camera
                                # old way to remove user
                                if object_id in self.users:
                                    if self.user_left_callback:
                                        self.callback_wrapper(self.user_left_callback, self.users[object_id], payload)
                                    del self.users[object_id]
                            elif (
                                HandLeft.object_type in object_id or HandRight.object_type in object_id
                            ):  # object is a hand/controller
                                if "dep" in obj.data:
                                    user_id = obj.data.dep
                                    if user_id in self.users:
                                        user = self.users[user_id]
                                        if obj in user.hands.values():
                                            if user.hand_remove_callback:
                                                self.callback_wrapper(user.hand_remove_callback, obj, payload)
                                            hand_key = (
                                                HandLeft.object_type
                                                if HandLeft.object_type in object_id
                                                else HandRight.object_type
                                            )
                                            del user.hands[hand_key]
                            elif self.delete_obj_callback:
                                self.callback_wrapper(self.delete_obj_callback, obj, payload)
                            Object.remove(obj)
                            span.add_event("Object delete.")

                            continue

                        else:  # create/update
                            obj.update_attributes(**payload)
                            span.add_event("Object attributes update.")

                    else:
                        self.telemetry.set_error("No message action!", span)

                    # call new message callback for all messages
                    if self.on_msg_callback:
                        if not event:
                            self.callback_wrapper(self.on_msg_callback, obj, payload)
                        else:
                            self.callback_wrapper(self.on_msg_callback, event, payload)

                    if object_type:
                        # run user_join_callback when user is found
                        if object_type == Camera.object_type:
                            if object_id not in self.users:
                                if object_id in self.all_objects:
                                    self.users[object_id] = obj
                                else:
                                    self.users[object_id] = Camera(**payload)
                                self.reset_private_objects(object_id)

                                if self.user_join_callback:
                                    self.callback_wrapper(self.user_join_callback, self.users[object_id], payload)

                        elif object_type == HandLeft.object_type or object_type == HandRight.object_type:
                            if "dep" in obj.data:
                                user_id = obj.data.dep
                                if user_id in self.users:
                                    user = self.users[user_id]
                                    if obj not in user.hands.values():
                                        user.hands[object_type] = obj
                                        obj.camera = user

                                        if user.hand_found_callback:
                                            self.callback_wrapper(user.hand_found_callback, obj, payload)

                    # if its an object the library has not seen before, call new object callback
                    elif object_id not in self.unspecified_object_ids and self.new_obj_callback:
                        self.callback_wrapper(self.new_obj_callback, obj, payload)
                        self.unspecified_object_ids.add(object_id)
                        span.add_event("New Object.")

                    span.add_event("Handle Msg Done.")

                except Exception as e:
                    self.telemetry.set_error(
                        f"Something went wrong!\n"
                        f"-----------------------------\n"
                        f"Source:\n{traceback.format_exc()}\n\n"
                        f"Exception occured when processing payload: {payload}\n"
                        f"-----------------------------\n"
                    )

    def callback_wrapper(self, func, arg, msg):
        """Checks for number of arguments for callback."""
        if len(signature(func).parameters) != 3:
            print("[DEPRECATED]", "Callbacks and handlers now take 3 arguments: (scene, obj/evt, msg)!")
            func(arg)
        else:
            func(self, arg, msg)

    def generate_custom_event(self, evt, action="clientEvent"):
        """Publishes an custom event. Could be user or library defined."""
        return self._publish(evt, action)

    def generate_click_event(self, obj: Object, type="mousedown", **kwargs):
        """Publishes an click event."""
        _type = type
        evt = Event(
            object_id=self.mqttc_id, type=_type, targetPosition=obj.data.position, target=obj.object_id, **kwargs
        )
        return self.generate_custom_event(evt, action="clientEvent")

    def manipulate_camera(self, cam, **kwargs):
        """Publishes a camera manipulation event."""
        if kwargs.get("position"):
            if isinstance(kwargs["position"], (list, tuple)):
                kwargs["position"] = Position(*kwargs["position"])
            elif isinstance(kwargs["position"], dict):
                kwargs["position"] = Position(**kwargs["position"])

        if kwargs.get("rotation"):
            if isinstance(kwargs["rotation"], (list, tuple)):
                kwargs["rotation"] = Rotation(*kwargs["rotation"])
            elif isinstance(kwargs["rotation"], dict):
                kwargs["rotation"] = Rotation(**kwargs["rotation"])

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id, type="camera-override", object_type=Camera.object_type, **kwargs)
        return self.generate_custom_event(evt, action="update")

    def look_at(self, cam, target):
        """Publishes a camera manipulation event."""
        if isinstance(target, tuple) or isinstance(target, list):
            target = Position(*target)
        elif isinstance(target, dict):
            target = Position(**target)
        elif isinstance(target, Object):
            target = target.object_id

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id, type="camera-override", object_type="look-at", target=target)
        return self.generate_custom_event(evt, action="update")

    def teleport_to_landmark(self, cam, target):
        """Publishes a camera manipulation event."""
        if isinstance(target, Object):
            target_id = target.object_id
        elif type(target) is str:
            target_id = target
        else:
            raise ValueError("target must be an ARENA object or string object_id")

        if isinstance(cam, Object):
            object_id = cam.object_id
        elif type(target) is str:
            object_id = cam
        else:
            raise ValueError("cam must be an ARENA object or string object_id")

        evt = Event(
            object_id=object_id, type="camera-override", object_type="teleport-to-landmark", landmarkObj=target_id
        )
        return self.generate_custom_event(evt, action="update")

    @property
    def all_objects(self):
        """Returns all the objects in a scene."""
        return Object.all_objects

    def get_private_objects(self, userid=None):
        """ Returns all private user objects."""
        if userid is not None:
            return Object.private_objects.get(userid, None)
        else:
            return Object.private_objects

    def reset_private_objects(self, userid):
        """Resets all private user objects."""
        Object.private_objects[userid] = {}

    def add_object(self, obj):
        """Public function to create an object."""
        if not isinstance(obj, Object):
            raise ValueError(f"Not a valid ARENA object to add to scene: {type(obj)}")
        # We have to set program_id here, as only scene has access to its userid
        if getattr(obj, "private", True):
            obj.program_id = self.userid
        res = self._publish(obj, "create")
        self.run_animations(obj)
        return res

    def add_objects(self, objs):
        """Public function to create multiple objects in a list."""
        for obj in objs:
            self.add_object(obj)
        return len(objs)

    def update_object(self, obj: Object, **kwargs):
        """Public function to update an object."""
        if kwargs:
            obj.update_attributes(**kwargs)

        # We have to update program_id here, as only scene has access to its userid
        if getattr(obj, "private", True):
            obj.program_id = self.userid

        # Check if any keys in delayed_prop_tasks are pending new animations
        # and cancel corresponding final update tasks or, if they are in
        # kwarg property updates, cancel the task as well as the animation
        if len(obj.delayed_prop_tasks) > 0:
            # If we have new animations that override a delayed task, cancel the task
            for anim in obj.animations:
                if anim.property in obj.delayed_prop_tasks:
                    task_to_cancel = obj.delayed_prop_tasks[anim.property]
                    task_to_cancel.cancel()
                    del task_to_cancel

            # If we have a property update that overrides a delayed task, cancel task
            # and cancel the animation
            for k in kwargs:
                if str(k) in obj.delayed_prop_tasks:
                    task_to_cancel = obj.delayed_prop_tasks[k]
                    task_to_cancel.cancel()
                    del task_to_cancel
                    obj.dispatch_animation(Animation(property=k, enabled=False, dur=0))

        res = self._publish(obj, "update")
        self.run_animations(obj)
        return res

    def update_objects(self, objs, **kwargs):
        """Public function to update multiple objects in a list."""
        for obj in objs:
            self.update_object(obj, **kwargs)
        return len(objs)

    def delete_object(self, obj):
        """Public function to delete an object."""
        payload = {"object_id": obj.object_id}
        Object.remove(obj)
        return self._publish(payload, "delete", custom_payload=True)

    def delete_user_objects(self, userid):
        """Deletes any private user objects."""
        if userid in Object.private_objects:
            for obj in Object.private_objects[userid].keys():
                Object.all_objects.pop(obj, None)
        del Object.private_objects[userid]

    def delete_program(self, obj):
        type=None
        try:
            type = obj.type
        except AttributeError:
            raise ValueError("obj must be of type Program!")
        if not type == Program.object_type: ValueError("obj must be of type Program!")

        # create runtime request (to delete program instance); TODO: Properly define a program instance/runtime request
        payload = {
            "object_id": str(uuid.uuid4()),
            "type": "req",
            "data": {
                "uuid": obj.object_id,
                "type": "module",
                "name": obj.data.name,
                "parent": obj.data.parent,
            },
        }
        Object.remove(obj)
        return self._publish(payload, "delete", custom_payload=True, publish_topic=PUBLISH_TOPICS.SCENE_PROGRAM)

    def delete_attributes(self, obj: Object, attributes=None):
        """Public function to delete a list of 'attributes' as a string[], updating each to null."""
        updated_data = {}
        for attr in attributes:
            obj.data[attr] = None  # remove from large internal storage
            updated_data[attr] = None  # remove from small external publish
        payload = {
            "object_id": obj.object_id,
            "type": obj.type,
            "persist": obj.persist,
            "data": updated_data,  # dashes handled from string array
        }
        return self._publish(payload, "update", custom_payload=True)

    def run_animations(self, obj: Object):
        """Runs all dispatched animations."""
        if isinstance(obj, Object):
            payload = {"object_id": obj.object_id, "type": obj.type, "data": {"object_type": obj.data.object_type}}
            if len(obj.animations) == 0:
                return
            for i, animation in enumerate(obj.animations):
                if isinstance(animation, AnimationMixer):
                    payload["data"][f"animation-mixer"] = vars(animation)
                else:
                    anim = vars(animation).copy()
                    payload["data"][f"animation__{anim['property']}"] = anim
                    Utils.dict_key_replace(anim, "start", "from")
                    Utils.dict_key_replace(anim, "end", "to")
                    self.create_delayed_task(obj, anim)
            obj.clear_animations()
            return self._publish(payload, "update", custom_payload=True)
        else:
            raise ValueError("obj must be of type Object!")

    def create_delayed_task(self, obj: Object, anim: Animation | AnimationMixer):
        """
        Creates a delayed task to push the end state of an animation after the expected
        duration. Uses async sleep to avoid blocking.
        :param obj: arena object to update.
        :param anim: Animation to run.
        :return: created async task.
        """

        async def _delayed_task():
            try:
                sleep_dur = (anim.get("dur", 0) + anim.get("delay", 0)) / 1000
                await asyncio.sleep(sleep_dur)  # this is in seconds
                final_state = anim["from"] if anim.get("dir", "normal") == "reverse" else anim["to"]
                obj.update_attributes(**{anim["property"]: final_state})
                self.update_object(obj)
                obj.delayed_prop_tasks.pop(anim["property"], None)
            except asyncio.CancelledError:
                self.telemetry.add_event(f"Animation end task cancelled for {obj.object_id}.{anim['property']}")

        if anim.get("loop", 0) or anim.get("enabled", True) is False:
            return None

        delayed_task = asyncio.create_task(_delayed_task())
        obj.delayed_prop_tasks[anim["property"]] = delayed_task
        delayed_task.object_id = obj.object_id
        return delayed_task

    def _publish(self, obj: Object, action, custom_payload=False, publish_topic=PUBLISH_TOPICS.SCENE_OBJECTS):
        """Publishes to mqtt broker with "action":action."""
        obj_type = None
        if "type" in obj:
            obj_type = obj["type"]

        with self.telemetry.start_publish_span(obj["object_id"], action, obj_type) as span:
            topic = publish_topic.substitute({**self.topicParams, **{"objectId": obj["object_id"]}})

            if publish_topic == PUBLISH_TOPICS.SCENE_OBJECTS:
                # self.can_publish_obj indicates if we can publish on the default publish_topic (PUBLISH_TOPICS.SCENE_OBJECTS)
                if not self.can_publish_obj:
                    self.telemetry.set_error(
                        f"ERROR: Publish failed! You do not have permission to publish to topic {topic} on {self.web_host}",
                        span,
                    )
                elif hasattr(obj, "_private_userid"):
                    topic = PUBLISH_TOPICS.SCENE_OBJECTS_PRIVATE.substitute(
                        {**self.topicParams, **{"objectId": obj['object_id'], "toUid": obj['_private_userid']}}
                    )

            d = datetime.utcnow().isoformat()[:-3] + "Z"

            if custom_payload:
                tmp_obj = obj
                tmp_obj["action"] = action
                tmp_obj["timestamp"] = d
                payload = json.dumps(tmp_obj)
            else:
                payload = obj.json(action=action, timestamp=d)

            # add warning to user that outgoing rotation is invalid if auto/manual units are incorrect
            if '"w": null' in payload:
                self.telemetry.set_error(
                    f"ERROR!! Publishing wire rotation data must be in Quaternion units, Euler conversion failed for payload: {payload}"
                )

            self.mqttc.publish(topic, payload, qos=0)
            if self.debug:
                self.telemetry.add_event(f"[publish] {topic} {payload}")
            return payload

    def get_persisted_obj(self, object_id):
        """Returns a dictionary for a persisted object.

        If object is known by arena-py, return local object, not persisted.
        """
        persist_obj = None
        if object_id in self.all_objects:
            persist_obj = self.all_objects[object_id]
            persist_obj.persist = True
        else:
            # pass token to persist
            data = self.auth.urlopen(url=f"{self.persist_url}/{object_id}", creds=True)
            persist_obj = json.loads(data)
            if len(persist_obj) > 0:
                persist_obj = persist_obj[0]

                obj_id = persist_obj.get("object_id", object_id)
                data = persist_obj.get("attributes", {})
                object_type = data.get("object_type")

                # special case for Program type
                if persist_obj.get("type") == Program.object_type:
                    object_type = Program.object_type

                if object_type != None:
                    obj_class = OBJECT_TYPE_MAP.get(object_type, Object)
                    persist_obj = obj_class(object_id=obj_id, data=data)
                    persist_obj.persist = True

        if not persist_obj:
            raise ValueError(f'object with object_id="{object_id}" not found in persist!')

        return persist_obj

    def get_persisted_objs(self):
        """Returns a dictionary of persisted objects.

        If object is known by arena-py, return our local object, not persisted.
        Silently fails/skip objects without object_id and object_type (except programs).
        Instantiates generic Object if object_type is given but unknown to arena-py.
        """
        objs = {}
        # pass token to persist
        data = self.auth.urlopen(url=self.persist_url, creds=True)
        output = json.loads(data)
        for obj in output:
            # note: no exception on missing fields
            object_id = obj.get("object_id")
            data = obj.get("attributes", {})
            object_type = data.get("object_type")

            # special case for Program type
            if obj.get("type") == Program.object_type:
                object_type = Program.object_type

            persisted_obj = None

            if object_id != None:
                if object_id in self.all_objects:
                    # note: object from our list (not from persist)
                    persisted_obj = self.all_objects[object_id]
                    persisted_obj.persist = True

                    # replace program object, if matches our program id
                    if object_type == Program.object_type:
                        if os.environ.get(PROGRAM_OBJECT_ID) == object_id:
                            print("PROGRAM_OBJECT_ID", object_id)
                            persisted_obj.persist = True
                            self.program = Program(object_id=object_id, **data)
                        else:
                            # dont persist program object if not ours
                            persisted_obj.persist = False

                elif object_type != None:
                    # note: instantiate even with empty attributes if object_type is unknown to arena-py
                    obj_class = OBJECT_TYPE_MAP.get(object_type, Object)
                    persisted_obj = obj_class(object_id=object_id, data=data)
                    persisted_obj.persist = True

                if persisted_obj is not None:
                    objs[object_id] = persisted_obj

        return objs

    def get_persisted_scene_option(self):
        """Returns a dictionary for scene-options."""
        # [TODO] wrap the output as a BaseObject
        scene_opts_url = f"{self.persist_url}?type=scene-options"
        # pass token to persist
        data = self.auth.urlopen(url=scene_opts_url, creds=True)
        output = json.loads(data)
        return output

    def get_writable_scenes(self):
        """Request list of scene names for logged in user account that user has publish permission for.

        :returns list: list of scenes.
        """
        return self.auth.get_writable_scenes(web_host=self.web_host)

    def get_user_list(self):
        """Returns a list of users."""
        return self.users.values()

    def get_rcv_pub_queue_len(self):
        """Return QueueStats object with receive and publish queue length."""
        return QueueStats(super().rcv_queue_len(), super().pub_queue_len())

    def run_info_update(self, run_info):
        """Callback when program stats are updated; publish program object update."""
        # Add run info to program data object and publish program object update
        run_info.add_program_info(self.program.data)
        self._publish(self.program, "update", publish_topic=PUBLISH_TOPICS.SCENE_PROGRAM)

    def upload_store_file(self, src_file_path, dest_file_path=None):
        """Upload a file to the filestore using the user's Google account.

        :param str src_file_path: Local path to the file to upload (required).
        :param str dest_file_path: Destination file path, can include dirs. Defaults to filename from src_file_path (optional).
        :return str: Url address of successful file upload location, or None if failed.
        """
        return self.auth.upload_store_file(
            web_host=self.web_host, scenename=self.scene, src_file_path=src_file_path, dest_file_path=dest_file_path
        )


class Arena(Scene):
    """
    Another name for Scene.
    """



from .arena_object import Object

class Image(Object):
    """
    Image object class to manage its properties in the ARENA: Display an image on a plane. See guidance to store paths under <https://docs.arenaxr.org/content/interface/filestore.html> ARENA File Store, CDN, or DropBox.

    :param float height: height Defaults to '1' (optional)
    :param int segmentsHeight: segments height Defaults to '1' (optional)
    :param int segmentsWidth: segments width Defaults to '1' (optional)
    :param str url: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    :param float width: width Defaults to '1' (optional)
    """
    object_type = "image"

    def __init__(self, url="", **kwargs):
        super().__init__(object_type=Image.object_type, url=url, **kwargs)



import json
import os
import re
import sys

from .arena_mqtt import ArenaMQTT
from .env import DEVICE, _get_env


class Device(ArenaMQTT):
    """
    Gives access to an ARENA device.
    Can create and execute various user-defined functions/tasks.

    :param str host: Hostname of the ARENA webserver (required).
    :param str realm: Reserved topic fork for future use (optional).
    :param str namespace: Username of authenticated user or other namespace (automatic).
    :param str device: The name of the device, without namespace (required).
    :param int network_latency_interval: Interval (in ms) to run network graph latency update. Default value is 10000 (10 secs). Ignore this parameter (optional).
    :param func on_msg_callback: Called on all MQTT messages received (optional).
    :param func end_program_callback: Called on MQTT disconnect (optional).
    :param bool debug: If true, print a log of all publish messages from this client (optional).
    """

    def __init__(
        self,
        host="arenaxr.org",
        realm="realm",
        network_latency_interval=10000,  # run network latency update every 10s
        on_msg_callback=None,
        end_program_callback=None,
        debug=False,
        **kwargs,
    ):

        if os.environ.get(DEVICE):
            self.device = _get_env(DEVICE)
            print(f"Using Device from 'DEVICE' env variable: {self.device}")
        elif "device" in kwargs and kwargs["device"]:
            if re.search("/", kwargs["device"]):
                sys.exit("Device argument (device) cannot include '/', aborting...")
            self.device = kwargs["device"]
            print(f"Using Device from 'device' input parameter: {self.device}")
        else:
            sys.exit("Device argument (device) is unspecified or None, aborting...")

        super().__init__(host, realm, network_latency_interval, on_msg_callback, end_program_callback, debug, **kwargs)
        print(f"Device topic ready: {self.realm}/d/{self.namespace}/{self.device}, mqtt_host={self.mqtt_host}")

    async def process_message(self):
        while True:
            msg = await self.msg_queue.get()
            # extract payload
            try:
                payload_str = msg.payload.decode("utf-8", "ignore")
                payload = json.loads(payload_str)
            except Exception as e:
                print("Malformed payload, ignoring:")
                print(e)
                return

    def publish(self, topic, payload_obj):
        """Publishes to mqtt broker."""
        payload = json.dumps(payload_obj)
        self.mqttc.publish(topic, payload, qos=0)
        if self.debug:
            print("[publish]", topic, payload)
        return payload

    def on_publish(self, client, userdata, mid):
        pass



# simple general purpose functions
import numpy as np


class Utils(object):
    @classmethod
    def tuple_to_string(cls, tup, sep=" "):
        """Turns a tuple into a string."""
        s = ""
        for val in tup:
            s += str(val) + sep
        return s.strip()

    @classmethod
    def agran(cls, float_num):
        """Reduces floating point numbers to ARENA granularity."""
        if isinstance(float_num, str):
            try:
                float_num = float(float_num)
            except:
                pass
        return round(float_num, 3)

    @classmethod
    def dict_key_replace(cls, d, key, new_key):
        """Replaces a key,val with a new key,val."""
        if key in d:
            ref = d[key]
            del d[key]
            d[new_key] = ref
        return d

    @classmethod
    def quat_to_matrix3(cls, rotq):
        x, y, z, w = rotq

        # Compute matrix elements
        xx, xy, xz = x * x, x * y, x * z
        yy, yz, zz = y * y, y * z, z * z
        wx, wy, wz = w * x, w * y, w * z

        # Rotation matrix
        rot_matrix = np.array(
            [
                [1 - 2 * (yy + zz), 2 * (xy - wz), 2 * (xz + wy)],
                [2 * (xy + wz), 1 - 2 * (xx + zz), 2 * (yz - wx)],
                [2 * (xz - wy), 2 * (yz + wx), 1 - 2 * (xx + yy)],
            ]
        )

        return rot_matrix

    @classmethod
    def matrix3_to_quat(cls, rotm):
        m00, m01, m02 = rotm[0, 0], rotm[0, 1], rotm[0, 2]
        m10, m11, m12 = rotm[1, 0], rotm[1, 1], rotm[1, 2]
        m20, m21, m22 = rotm[2, 0], rotm[2, 1], rotm[2, 2]

        # Compute quaternion components
        w = np.sqrt(max(0, 1 + m00 + m11 + m22)) / 2
        x = np.sqrt(max(0, 1 + m00 - m11 - m22)) / 2
        y = np.sqrt(max(0, 1 - m00 + m11 - m22)) / 2
        z = np.sqrt(max(0, 1 - m00 - m11 + m22)) / 2

        x = np.copysign(x, m21 - m12)
        y = np.copysign(y, m02 - m20)
        z = np.copysign(z, m10 - m01)

        return np.array([x, y, z, w])

    @classmethod
    def pose_to_matrix4(cls, pos, rotq, scale=(1, 1, 1)):  # Def arg not mutated
        mat = np.identity(4)
        mat[0:3, 0:3] = cls.quat_to_matrix3([rotq.x, rotq.y, rotq.z, rotq.w])
        mat[0:3, 3] = [pos.x, pos.y, pos.z]
        if scale != (1, 1, 1):
            scale_mat = np.identity(4)
            scale_mat[0:3, 0:3] = np.diag(scale)
            mat = mat @ scale_mat
        return mat

    @classmethod
    def matrix4_to_pose(cls, mat):
        pos = mat[0:3, 3]
        rotq = cls.matrix3_to_quat(mat[0:3, 0:3])
        scale = np.sqrt(np.sum(mat[0:3, 0:3] ** 2, axis=0))
        return pos, rotq, scale

    @classmethod
    def get_world_pose(cls, obj, scene):
        current_obj = obj
        matrices = []
        while "parent" in current_obj.data:
            current_matrix = cls.pose_to_matrix4(
                current_obj.data.position,
                current_obj.data.rotation.quaternion,
                current_obj.data.scale.array,
            )
            matrices = [current_matrix] + matrices  # prepend
            current_obj = scene.all_objects[current_obj.data.parent]
        final_matrix = np.identity(4)
        for matrix in matrices:
            final_matrix = final_matrix @ matrix
        return cls.matrix4_to_pose(final_matrix)



from string import Template
from types import SimpleNamespace

TOPIC_TYPES = SimpleNamespace(**{
    'SCENE': 's',
    'PROC': 'p',
    'DEVICE': 'd',
})

DEVICE_TOPIC_TOKENS = SimpleNamespace(**{
    'REALM': 0,
    'TYPE': 1,
    'DEVICE_NAME': 2,
    'UUID': 3,
})

TOPIC_TOKENS = SimpleNamespace(**{
    'REALM': 0,
    'TYPE': 1,
    'NAMESPACE': 2,
    'SCENENAME': 3,
    'SCENE_MSGTYPE': 4,
    'USER_CLIENT': 5,
    'UUID': 6,
    'TO_UID': 7,
})

SCENE_MSGTYPES = SimpleNamespace(**{
    'PRESENCE': 'x',
    'CHAT': 'c',
    'USER': 'u',
    'OBJECTS': 'o',
    'RENDER': 'r',
    'ENV': 'e',
    'PROGRAM': 'p',
    'DEBUG': 'd',
})

SUBSCRIBE_TOPICS = SimpleNamespace(**{
    'NETWORK':                '$NETWORK',
    'DEVICE':                 Template('${realm}/d/${nameSpace}/${deviceName}/#'),  # All client placeholder
    'SCENE_PUBLIC':           Template('${realm}/s/${nameSpace}/${sceneName}/+/+/+'),
    'SCENE_PUBLIC_SELF':      Template('${realm}/s/${nameSpace}/${sceneName}/+/${userClient}/+'),
    'SCENE_PRIVATE':          Template('${realm}/s/${nameSpace}/${sceneName}/+/+/+/${idTag}/#'),
    'SCENE_ENV_PRIVATE':      Template('${realm}/s/${nameSpace}/${sceneName}/e/+/+/${idTag}/#'),
})

PUBLISH_TOPICS = SimpleNamespace(**{
    'NETWORK_LATENCY':        '$NETWORK/latency',
    'DEVICE':                 Template('${realm}/d/${nameSpace}/${deviceName}/${idTag}'),
    'SCENE_PRESENCE':         Template('${realm}/s/${nameSpace}/${sceneName}/x/${userClient}/${idTag}'),
    'SCENE_PRESENCE_PRIVATE': Template('${realm}/s/${nameSpace}/${sceneName}/x/${userClient}/${idTag}/${toUid}'),
    'SCENE_CHAT':             Template('${realm}/s/${nameSpace}/${sceneName}/c/${userClient}/${idTag}'),
    'SCENE_CHAT_PRIVATE':     Template('${realm}/s/${nameSpace}/${sceneName}/c/${userClient}/${idTag}/${toUid}'),
    'SCENE_USER':             Template('${realm}/s/${nameSpace}/${sceneName}/u/${userClient}/${userObj}'),
    'SCENE_USER_PRIVATE':     Template('${realm}/s/${nameSpace}/${sceneName}/u/${userClient}/${userObj}/${toUid}'),  # Need to add face_ privs
    'SCENE_OBJECTS':          Template('${realm}/s/${nameSpace}/${sceneName}/o/${userClient}/${objectId}'),  # All client placeholder
    'SCENE_OBJECTS_PRIVATE':  Template('${realm}/s/${nameSpace}/${sceneName}/o/${userClient}/${objectId}/${toUid}'),
    'SCENE_RENDER':           Template('${realm}/s/${nameSpace}/${sceneName}/r/${userClient}/${idTag}'),
    'SCENE_RENDER_PRIVATE':   Template('${realm}/s/${nameSpace}/${sceneName}/r/${userClient}/${idTag}/-'),  # To avoid unpriv sub
    'SCENE_ENV':              Template('${realm}/s/${nameSpace}/${sceneName}/e/${userClient}/${idTag}'),
    'SCENE_ENV_PRIVATE':      Template('${realm}/s/${nameSpace}/${sceneName}/e/${userClient}/${idTag}/-'),  # To avoid unpriv sub
    'SCENE_PROGRAM':          Template('${realm}/s/${nameSpace}/${sceneName}/p/${userClient}/${idTag}'),
    'SCENE_PROGRAM_PRIVATE':  Template('${realm}/s/${nameSpace}/${sceneName}/p/${userClient}/${idTag}/${toUid}'),
    'SCENE_DEBUG':            Template('${realm}/s/${nameSpace}/${sceneName}/d/${userClient}/${idTag}/-'),  # To avoid unpriv sub
})



from .attribute import Attribute

class VideoControl(Attribute):
    """
    VideoControl attribute class to manage its properties in the ARENA: Adds a video to an entity and controls its playback.
    Usage: `video_control=VideoControl(...)`

    :param bool anyone_clicks: Responds to clicks from any user. Defaults to 'True' (optional)
    :param bool autoplay: Video starts playing automatically. Defaults to 'False' (optional)
    :param bool cleanup: Automatically remove HTML5 video and img assets from DOM on object removal. Defaults to 'True' (optional)
    :param str frame_object: URL of a thumbnail image, e.g. 'store/users/wiselab/images/conix-face-white.jpg'. (optional)
    :param bool video_loop: Video automatically loops. Defaults to 'True' (optional)
    :param str video_object: Name of object where to put the video, e.g. 'square_vid6'. (optional)
    :param str video_path: URL of the video file, e.g. 'store/users/wiselab/videos/kungfu.mp4'. (optional)
    :param float volume: Video sound volume. Defaults to '1' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .arena_object import Object


class ObjModel(Object):
    """
    ObjModel object class to manage its properties in the ARENA: Loads a 3D model and material using a Wavefront (.OBJ) file and a .MTL file. See guidance to store paths under <https://docs.arenaxr.org/content/interface/filestore.html> ARENA File Store, CDN, or DropBox.

    :param str mtl: Url pointing to a .MTL file. Optional if you wish to use the material component instead. (optional)
    :param str obj: Url pointing to a .OBJ file. Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    """
    object_type = "obj-model"

    def __init__(self, **kwargs):
        super().__init__(object_type=ObjModel.object_type, **kwargs)



from deprecated import deprecated

from .arena_object import Object


class Light(Object):
    """
    Light object class to manage its properties in the ARENA: A light. More properties at <https://aframe.io/docs/1.5.0/components/light.html> A-Frame Light.

    :param float angle: Maximum extent of spot light from its direction (in degrees). Requires type:spot. Defaults to '60' (optional)
    :param bool castShadow: Whether this light casts shadows on the scene. Defaults to 'False' (optional)
    :param str color: Light color. For 'hemisphere', light color from above. Defaults to '#ffffff' (optional)
    :param float decay: Amount the light dims along the distance of the light. Requires type:point or type:spot. Defaults to '1' (optional)
    :param float distance: Distance where intensity becomes 0. If distance is 0, then the point light does not decay with distance. Requires type:point or type:spot. Defaults to '0' (optional)
    :param str envMap: Cube Map to load. Defaults to 'None' (optional)
    :param str groundColor: Light color from below. Requires type:hemisphere. Defaults to '#ffffff' (optional)
    :param float intensity: Amount of light provided. Defaults to '1' (optional)
    :param dict light: ***DEPRECATED**: data.light.[property] is **deprecated**, use object_type: light and data.[property] instead.* (deprecated)
    :param float penumbra: Percent of the spotlight cone that is attenuated due to penumbra. Requires type:spot. Defaults to '0' (optional)
    :param float shadowBias: Offset depth when deciding whether a surface is in shadow. Tiny adjustments here (in the order of +/-0.0001) may reduce artifacts in shadows. Defaults to '0' (optional)
    :param float shadowCameraBottom: Bottom plane of shadow camera frustum. Requires type:directional. Defaults to '-5' (optional)
    :param float shadowCameraFar: Far plane of shadow camera frustum. Defaults to '500' (optional)
    :param float shadowCameraFov: Shadow camera's FOV. Requires type:point or spot. Defaults to '50' (optional)
    :param float shadowCameraLeft: Left plane of shadow camera frustum. Requires type:directional. Defaults to '-5' (optional)
    :param float shadowCameraNear: Near plane of shadow camera frustum. Defaults to '0.5' (optional)
    :param float shadowCameraRight: Right plane of shadow camera frustum. Requires type:directional. Defaults to '5' (optional)
    :param float shadowCameraTop: Top plane of shadow camera frustum. Requires type:directional. Defaults to '5' (optional)
    :param bool shadowCameraVisible: Displays a visual aid showing the shadow camera's position and frustum. This is the light's view of the scene, used to project shadows. Defaults to 'False' (optional)
    :param float shadowMapHeight: Shadow map's vertical resolution. Larger shadow maps display more crisp shadows, at the cost of performance. Defaults to '512' (optional)
    :param float shadowMapWidth: Shadow map's horizontal resolution. Defaults to '512' (optional)
    :param float shadowRadius: shadowRadius (castShadow=true) Defaults to '1' (optional)
    :param str target: Id of element the spot should point to. Set to null to transform spotlight by orientation, pointing to it's -Z axis. Requires type:spot. Defaults to 'None' (optional)
    :param str type: The type of light, or what shape the light should take. Allows [ambient, directional, hemisphere, point, spot] Defaults to 'directional' (optional)
    """
    object_type = "light"

    def __init__(self, **kwargs):
        super().__init__(object_type=Light.object_type, **kwargs)

    @property
    @deprecated("data.light.[property] is deprecated, use object_type: light and data.[property] instead.")
    def light(self):
        return None

    @light.setter
    @deprecated("data.light.[property] is deprecated, use object_type: light and data.[property] instead.")
    def light(self, value):
        return



from .arena_object import Object


class UrdfModel(Object):
    """
    UrdfModel object class to manage its properties in the ARENA: Load a URDF model. See guidance to store paths under <https://docs.arenaxr.org/content/interface/filestore.html> ARENA File Store, CDN, or DropBox.

    :param str joints: Set joint values (in degrees) in the form 'JointName1: ValueInDegrees1, JointName2: ValueInDegrees2, ...'.  (optional)
    :param str url: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    :param str urlBase: Base path for xacro/urdf package includes. This must be an absolute path with no trailing slash, e.g. '/store/users/username/robot' (optional)
    """
    object_type = "urdf-model"

    def __init__(self, **kwargs):
        super().__init__(object_type=UrdfModel.object_type, **kwargs)



from .attribute import Attribute


class GotoLandmark(Attribute):
    """
    GotoLandmark attribute class to manage its properties in the ARENA: Teleports user to the landmark with the given name. Requires click-listener.
    Usage: `goto_landmark=GotoLandmark(...)`

    :param str landmark: Id of landmark to teleport to. Defaults to '' (optional)
    :param str on: Event to listen 'on'. Allows [mousedown, mouseup] Defaults to 'mousedown' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class Multisrc(Attribute):
    """
    Multisrc attribute class to manage its properties in the ARENA: Define multiple visual sources applied to an object.
    Usage: `multisrc=Multisrc(...)`

    :param str srcs: A comma-delimited list if URIs, e.g. 'side1.png, side2.png, side3.png, side4.png, side5.png, side6.png' (required). (optional)
    :param str srcspath: URI, relative or full path of a directory containing srcs, e.g. 'store/users/wiselab/images/dice/' (required). (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from ..utils import Utils
from .attribute import Attribute
from .position import Position

class Impulse(Attribute):
    """
    Impulse attribute class to manage its properties in the ARENA: Apply an impulse to an object to set it in motion. This happens in conjunction with an event. Requires click-listener and physics.
    Usage: `impulse=Impulse(...)`

    :param dict force: Impulse vector. Defaults to '{'x': 1, 'y': 1, 'z': 1}' (optional)
    :param str on: Event to listen 'on'. Allows [mousedown, mouseup] Defaults to 'mousedown' (optional)
    :param dict position: World position. Defaults to '{'x': 1, 'y': 1, 'z': 1}' (optional)
    """
    def __init__(self, on="mousedown", force=Position(0,0,0), position=Position(0,0,0), **kwargs):
        if isinstance(force, Position):
            force = force.to_str()
        elif isinstance(force, tuple) or isinstance(force, list):
            force = Utils.tuple_to_string(force)

        if isinstance(position, Position):
            position = position.to_str()
        elif isinstance(position, tuple) or isinstance(position, list):
            position = Utils.tuple_to_string(position)

        super().__init__(on=on, force=force, position=position, **kwargs)



"""
auth.py - Authentication methods for accessing the ARENA.
"""

import datetime
import json
import os
import re
import ssl
import sys
import time
import webbrowser
from pathlib import Path
from urllib import parse, request
from urllib.error import HTTPError, URLError
from urllib.parse import urlsplit

import jwt
import requests
from google.auth import jwt as gJWT
from google_auth_oauthlib.flow import InstalledAppFlow

from .utils import topic_matches_sub

_gauth_file = ".arena_google_auth"
_mqtt_token_file = ".arena_mqtt_auth"
_arena_user_dir = f"{str(Path.home())}/.arena"
_local_mqtt_path = f"{_mqtt_token_file}"


class ArenaAuth:

    _scopes = [
        "openid",
        "https://www.googleapis.com/auth/userinfo.profile",
        "https://www.googleapis.com/auth/userinfo.email",
    ]

    def __init__(self):
        self._csrftoken = None
        self._mqtt_token = None
        self._id_token = None
        self._user_info = None
        self._store_token = None

    def authenticate_user(self, web_host, headless):
        """
        Begins authentication flow, getting Google auth, opening web browser if
        needed, getting username and state from ARENA server.

        :param str web_host: The hostname of the ARENA webserver.
        :return: Username from arena-account, or None.
        """
        print("Signing in to the ARENA...")
        try:
            # test for valid browser before starting browser-required auth-flow
            webbrowser.get()
        except webbrowser.Error as err:
            headless = True
            print(f"Console-only environment detected. {err} ")

        if headless:
            gauth_json = self._get_gauthid_limited(web_host)
        else:
            gauth_json = self._get_gauthid_desktop(web_host)
        gauth = json.loads(gauth_json)
        if "installed" not in gauth or "client_id" not in gauth["installed"]:
            return None
        creds = None
        refresh_token = None
        scene_auth_dir = self._get_scene_auth_path(web_host)
        scene_gauth_path = f"{scene_auth_dir}/{_gauth_file}"

        if not os.path.exists(scene_auth_dir):
            os.makedirs(scene_auth_dir)

        # store the user's access and refresh tokens
        if os.path.exists(scene_gauth_path):
            try:
                with open(scene_gauth_path, "r", encoding="utf-8") as token:
                    creds = json.load(token)
            except (json.JSONDecodeError, UnicodeDecodeError):
                creds = None  # bad/old storage format

            if creds and "id_token" in creds:
                id_claims = gJWT.decode(creds["id_token"], verify=False)
                # for reuse, client_id must still match
                if id_claims["aud"] != gauth["installed"]["client_id"]:
                    creds = None  # switched auth systems
            if creds and "refresh_token" in creds:
                refresh_token = creds["refresh_token"]
            if creds and id_claims:
                exp = float(id_claims["exp"])
                if exp <= time.time():
                    creds = None  # expired token

        if creds:
            print("Using cached Google authentication.")
        else:
            if refresh_token:
                print("Requesting refreshed Google authentication.")
                creds_jstr = self._run_gauth_token_refresh(
                    client_id=gauth["installed"]["client_id"],
                    client_secret=gauth["installed"]["client_secret"],
                    refresh_token=refresh_token,
                )
                creds = json.loads(creds_jstr)
            else:
                # if no credentials available, let the user log in.
                if headless:
                    # limited input device auth flow for local client
                    print("Requesting new device Google authentication.")
                    creds_jstr = self._run_gauth_device_flow(
                        client_id=gauth["installed"]["client_id"],
                        client_secret=gauth["installed"]["client_secret"],
                    )
                    creds = json.loads(creds_jstr)
                else:
                    # automated browser flow for local client
                    print("Requesting new browser Google authentication.")
                    flow = InstalledAppFlow.from_client_config(json.loads(gauth_json), self._scopes)
                    credentials = flow.run_local_server(port=0)
                    creds_jstr = credentials.to_json()
                    creds = json.loads(creds_jstr)
                    creds["id_token"] = credentials.id_token

            with open(scene_gauth_path, "w", encoding="utf-8") as token:
                # save the credentials for the next run
                json.dump(creds, token)
            os.chmod(scene_gauth_path, 0o600)  # set user-only perms.

        username = None
        self._id_token = creds["id_token"]
        user_info = self._get_user_state(web_host)
        self._user_info = json.loads(user_info)
        if "authenticated" in self._user_info and "username" in self._user_info:
            username = self._user_info["username"]
        id_claims = gJWT.decode(creds["id_token"], verify=False)
        print(f"Authenticated Google account: {id_claims['email']}")
        return username

    def _run_gauth_token_refresh(self, client_id, client_secret, refresh_token):
        refresh_resp, status, url = self._get_gauth_refresh_token(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
        if 200 <= status <= 299:  # success
            return refresh_resp
        else:  # error
            print(f"HTTP error {status}: {url}\n{refresh_resp}")
            sys.exit("Terminating...")

    def _run_gauth_device_flow(self, client_id, client_secret):
        device_resp, status, url = self._get_gauth_device_code(client_id=client_id)
        if 200 > status > 299:  # error
            print(f"HTTP error {status}: {url}\n{device_resp}")
            sys.exit("Terminating...")
        # render user code/link and poll for OOB response
        device = json.loads(device_resp)
        print(f"1. Go to this page on any device: {device['verification_url']}")
        print(f"2. Enter this code on that page: {device['user_code']}")
        exp = time.time() + device["expires_in"]
        delay = device["interval"]

        next_time = time.time() + delay
        while True:
            time.sleep(max(0, next_time - time.time()))

            if time.time() > exp:
                print(f"Device auth request expired after {delay/60} minutes.")
                sys.exit("Terminating...")

            access_resp, status, url = self._get_gauth_device_token(
                client_id=client_id,
                client_secret=client_secret,
                device_code=device["device_code"],
            )
            if 200 <= status <= 299:  # success
                return access_resp
            if status == 428:  # awaiting remote user code and approval
                # skip tasks if we are behind schedule:
                next_time += (time.time() - next_time) // delay * delay + delay
            else:  # error
                print(f"HTTP error {status}: {url}\n{access_resp}")
                sys.exit("Terminating...")

    def authenticate_scene(self, web_host, realm, scene, username, video=False, env=False):
        """End authentication flow, requesting permissions may change by owner
        or admin, for now, get a fresh mqtt_token each time.

        :param str web_host: The hostname of the ARENA webserver.
        :param str realm: The topic realm name.
        :param str scene: The namespace/scene name combination.
        :param str username: The ARENA username for the user.
        :param bool video: If Jitsi video conference is requested.
        :return: username and mqtt_token from arena-account.
        """
        scene_auth_dir = self._get_scene_auth_path(web_host)
        scene_mqtt_path = f"{scene_auth_dir}/{_mqtt_token_file}"

        print("Using remote-authenticated MQTT token.")
        mqtt_json = self._get_mqtt_token(web_host, realm, scene, username, video, env)
        print(mqtt_json)
        # save mqtt_token
        with open(scene_mqtt_path, "w", encoding="utf-8") as d:
            d.write(mqtt_json)
        os.chmod(scene_mqtt_path, 0o600)  # set user-only perms.

        self._mqtt_token = json.loads(mqtt_json)
        self._log_token()
        return self._mqtt_token

    def authenticate_device(self, web_host):
        """
        Check for device mqtt_token, ask for a missing one, and save to local memory.
        """
        device_auth_dir = self._get_device_auth_path(web_host)
        device_mqtt_path = f"{device_auth_dir}/{_mqtt_token_file}"
        # check token expiration
        _remove_credentials(device_auth_dir, expire=True)
        # load device token if valid
        if os.path.exists(device_mqtt_path):
            print("Using user long-term device MQTT token.")
            with open(device_mqtt_path, "r", encoding="utf8") as f:
                mqtt_json = f.read()
        else:
            if not os.path.exists(device_auth_dir):
                os.makedirs(device_auth_dir)
            print(f"Generate a token for this device at https://{web_host}/user/v2/profile")
            mqtt_json = input("Paste auth MQTT full JSON here for this device: ")
            # save mqtt_token
            with open(device_mqtt_path, "w", encoding="utf-8") as d:
                d.write(mqtt_json)
            os.chmod(device_mqtt_path, 0o600)  # set user-only perms.

        self._mqtt_token = json.loads(mqtt_json)
        self._log_token()
        return self._mqtt_token

    def has_publish_rights(self, token, topic):
        """Check the MQTT token for permission to publish to topic."""
        tok = jwt.decode(token, options={"verify_signature": False})
        for pub in tok["publ"]:
            if topic_matches_sub(pub.strip(), topic):
                return True
        return False

    def _get_scene_auth_path(self, web_host):
        return f"{_arena_user_dir}/python/{web_host}/s"

    def _get_device_auth_path(self, web_host):
        return f"{_arena_user_dir}/python/{web_host}/d"

    def get_writable_scenes(self, web_host):
        """Request list of scene names for logged in user that user has publish permission for.

        :param str web_host: The hostname of the ARENA webserver.
        :return: list of scenes.
        """
        my_scenes = self._get_my_scenes(web_host)
        return json.loads(my_scenes)

    def _log_token(self):
        """
        Update user with token in use.
        """
        username = None
        if "username" in self._mqtt_token:
            username = self._mqtt_token["username"]
        print(f"ARENA Token Username: {username}")

        now = time.time()
        tok = jwt.decode(self._mqtt_token["token"], options={"verify_signature": False})
        exp = float(tok["exp"])
        delta = exp - now
        dur_str = str(datetime.timedelta(milliseconds=delta * 1000))
        print(f"ARENA Token valid for: {dur_str}h")

    def store_environment_auth(self, username, token):
        """
        Keep a copy of the token in local memory for urlopen and other tasks.
        """
        if username and token:
            print("Using environment MQTT token.")
            self._mqtt_token = {"username": username, "token": token}
            self._log_token()

    def check_local_auth(self):
        """
        Check for local mqtt_token and save to local memory.
        """
        # TODO: remove local check after ARTS supports mqtt_token passing
        # 4 Oct 2021 remove deprecated user home creds path
        _remove_credentials(str(Path.home()))
        # check token expiration
        _remove_credentials(str(Path.cwd()), expire=True)
        # load local token if valid
        if os.path.exists(_local_mqtt_path):
            print("Using local MQTT token.")
            with open(_local_mqtt_path, "r", encoding="utf8") as f:
                mqtt_json = f.read()
            self._mqtt_token = json.loads(mqtt_json)
            self._log_token()
            return self._mqtt_token
        return None

    def _encode_params(self, params):
        query_string = parse.urlencode(params)
        data = query_string.encode("ascii")
        return data

    def _get_csrftoken(self, web_host):
        # get the csrftoken for django
        csrf_url = f"https://{web_host}/user/v2/login"
        client = requests.session()
        client.get(csrf_url, verify=self.verify(web_host))  # sets cookie
        if "csrftoken" in client.cookies:
            self._csrftoken = client.cookies["csrftoken"]
        elif "csrf" in client.cookies:
            self._csrftoken = client.cookies["csrf"]
        else:
            self._csrftoken = None
        return self._csrftoken

    def _get_gauthid_desktop(self, web_host):
        url = f"https://{web_host}/conf/gauth.json"
        return self.urlopen(url)

    def _get_gauthid_limited(self, web_host):
        url = f"https://{web_host}/conf/gauth-device.json"
        return self.urlopen(url)

    def _get_gauth_device_code(self, client_id):
        url = "https://oauth2.googleapis.com/device/code"
        params = {
            "client_id": client_id,
            "scope": "email profile",
        }
        body, status = self.urlopen_def(url, data=self._encode_params(params))
        return body, status, url

    def _get_gauth_device_token(self, client_id, client_secret, device_code):
        url = "https://oauth2.googleapis.com/token"
        params = {
            "client_id": client_id,
            "client_secret": client_secret,
            "device_code": device_code,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        }
        body, status = self.urlopen_def(url, data=self._encode_params(params))
        return body, status, url

    def _get_gauth_refresh_token(self, client_id, client_secret, refresh_token):
        url = "https://oauth2.googleapis.com/token"
        params = {
            "client_id": client_id,
            "client_secret": client_secret,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
        }
        body, status = self.urlopen_def(url, data=self._encode_params(params))
        return body, status, url

    def _confirm_gauth(self):
        if not self._id_token:
            raise IOError(
                "Google auth is required. Remove manual .arena_mqtt_token or env ARENA_PASSWORD. Headless auth options are available using `scene=Scene(..., headless=True).`"
            )

    def _get_my_scenes(self, web_host):
        self._confirm_gauth()
        url = f"https://{web_host}/user/v2/my_scenes"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(web_host)
        params = {"id_token": self._id_token}
        return self.urlopen(url, data=self._encode_params(params), csrf=self._csrftoken)

    def _get_user_state(self, web_host):
        self._confirm_gauth()
        url = f"https://{web_host}/user/v2/user_state"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(web_host)
        params = {"id_token": self._id_token}
        return self.urlopen(url, data=self._encode_params(params), csrf=self._csrftoken)

    def _get_store_login(self, web_host):
        self._confirm_gauth()
        url = f"https://{web_host}/user/v2/storelogin"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(web_host)
        params = {"id_token": self._id_token}
        return self.urlopen(url, data=self._encode_params(params), csrf=self._csrftoken)

    def upload_store_file(self, web_host, scenename, src_file_path, dest_file_path=None):
        """Upload a source file to the user's file store space. Google authentication is required.

        :param str web_host: The hostname of the ARENA webserver.
        :param str scenename: The scene name/id.
        :param str src_file_path: Local path to the file to upload (required).
        :param str dest_file_path: Destination file path, can include dirs. Defaults to filename from src_file_path (optional).
        :return str: Url address of successful file upload location, or None if failed.
        """
        self._confirm_gauth()
        # request FS login if this is the first time.
        if not self._store_token:
            self._get_store_login(web_host)
            if not self._store_token:
                raise IOError("Filestore login failed!")

        # send file to filestore
        if not dest_file_path:
            dest_file_path = Path(src_file_path).name
        safe_file_path = re.sub(r"/(\W+)/gi", "-", dest_file_path)
        if self._user_info["is_staff"]:
            store_res_prefix = f"users/{self._user_info['username']}/"
        else:
            store_res_prefix = ""
        user_file_path = f"scenes/{scenename}/{safe_file_path}"
        store_res_path = f"{store_res_prefix}{user_file_path}"
        store_ext_path = f"store/users/{self._user_info['username']}/{user_file_path}"
        url = f"https://{web_host}/storemng/api/resources/{store_res_path}?override=true"
        headers = {
            "Content-Length": os.stat(src_file_path).st_size,
            "X-Auth": self._store_token,
        }
        print(f"Uploading {src_file_path}....")
        with open(src_file_path, "rb") as f:
            body = self.urlopen(url, data=f, headers=headers)
        if body:
            print("Upload DONE!")
            return f"https://{web_host}/{store_ext_path}"
        else:
            raise IOError(f"Filestore upload failed! Dest: {dest_file_path}")

    def _get_mqtt_token(self, web_host, realm, scene, username, video, env):
        self._confirm_gauth()
        url = f"https://{web_host}/user/v2/mqtt_auth"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(web_host)
        params = {
            "id_auth": "google-installed",
            "username": username,
            "id_token": self._id_token,
            "client": "py",
            "realm": realm,
            "scene": scene,
            "userid": "true",
        }
        if video:
            params["camid"] = "true"
        if env:
            params["environmentid"] = "true"
        return self.urlopen(url, data=self._encode_params(params), csrf=self._csrftoken)

    def verify(self, web_host):
        return web_host != "localhost"

    def urlopen_def(self, url, data=None):
        """urlopen default is for non-ARENA URL connections.
        :param str url: the url to POST/GET.
        :param str data: None for GET, add params for POST.
        """
        body, status = None, None
        try:
            req = request.Request(url)
            with request.urlopen(req, data=data) as f:
                status = f.status
                body = f.read().decode("utf-8")
        except HTTPError as err:
            # do not log errors, allow consumer to decide
            status = err.code
            body = err.read().decode("utf-8")
        except (
            requests.exceptions.ConnectionError,
            ConnectionError,
            URLError,
        ) as err:
            print(f"{err}: {url}")

        return body, status

    def urlopen(self, url, data=None, headers=None, creds=False, csrf=None):
        """urlopen is for ARENA URL connections.
        :param str url: the url to POST/GET.
        :param str data: None for GET, add params for POST.
        :param bool creds: True to pass the MQTT token as a cookie.
        :param str csrf: The csrftoken.
        """
        urlparts = urlsplit(url)
        body = None
        try:
            req = request.Request(url)
            if creds:
                req.add_header("Cookie", f"mqtt_token={self._mqtt_token['token']}")
            if csrf:
                req.add_header("Cookie", f"csrftoken={csrf}")
                req.add_header("X-CSRFToken", csrf)
            if headers:
                for k in headers:
                    req.add_header(k, headers[k])
            if self.verify(urlparts.netloc):
                with request.urlopen(req, data=data) as f:
                    body = f.read().decode("utf-8")
                    cookies = f.info().get_all("Set-Cookie")
            else:
                context = ssl.create_default_context()
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
                with request.urlopen(req, data=data, context=context) as f:
                    body = f.read().decode("utf-8")
                    cookies = f.info().get_all("Set-Cookie")
            if cookies:
                for cookie in cookies:
                    if "auth=" in cookie:
                        for m in re.finditer(r"(^| )auth=([^;]+)", cookie):
                            self._store_token = m.group(2)
            return body
        except HTTPError as err:
            print(f"{err}: {url}")
            body = err.read().decode("utf-8")
            if "<!doctype html>" not in body.lower():
                print(body)  # show additional errors in response if present
            if err.code in (401, 403):
                # user not authorized on website yet, they don"t have an ARENA username
                us = urlsplit(url)
                base_url = f"{us.scheme}://{us.netloc}"
                print(f"Do you have a valid ARENA account on {base_url}?")
                print(f"Create an account in a web browser at: {base_url}/user")
            sys.exit("Terminating...")
        except (
            requests.exceptions.ConnectionError,
            ConnectionError,
            URLError,
        ) as err:
            print(f"{err}: {url}")
            sys.exit("Terminating...")


def signout():
    auth_files = [_mqtt_token_file, _gauth_file]
    for root, dirs, files in os.walk(_arena_user_dir):
        if any(map(lambda v: v in auth_files, files)):
            _remove_credentials(root)
    if os.path.exists(_local_mqtt_path):
        _remove_credentials(_local_mqtt_path)
    print("Signed out of the ARENA.")


def _print_mqtt_token(storage_str, mqtt_claims):
    print("\nARENA MQTT/Video Permissions")
    print("----------------------")
    exp_str = time.strftime("%c", time.localtime(mqtt_claims["exp"]))
    print(f"Expires: {exp_str}")
    print(f"Storage: {storage_str}")
    print(f"{json.dumps(mqtt_claims, indent=4)}")


def permissions():
    mqtt_claims = None
    # env storage auth
    if os.environ.get("ARENA_USERNAME") and os.environ.get("ARENA_PASSWORD"):
        mqtt_token = os.environ["ARENA_PASSWORD"]
        mqtt_claims = jwt.decode(mqtt_token["token"], options={"verify_signature": False})
        _print_mqtt_token("environment variable 'ARENA_PASSWORD'", mqtt_claims)
    # file storage auth
    token_paths = []
    for root, dirs, files in os.walk(_arena_user_dir):
        if _mqtt_token_file in files:
            token_paths.append(os.path.join(root, _mqtt_token_file))
    if os.path.exists(_local_mqtt_path):
        token_paths.append(_local_mqtt_path)
    for mqtt_path in token_paths:
        with open(mqtt_path, "r", encoding="utf8") as f:
            mqtt_json = f.read()
        try:
            mqtt_token = json.loads(mqtt_json)
        except json.decoder.JSONDecodeError as err:
            print(f"{err}, {mqtt_path}")
            continue
        mqtt_claims = jwt.decode(mqtt_token["token"], options={"verify_signature": False})
        _print_mqtt_token(mqtt_path, mqtt_claims)
    # no permissions
    if not mqtt_claims:
        print("Not signed into the ARENA.")


def _remove_credentials(cred_dir, expire=False):
    """
    Helper to remove credentials in path with expiration option.
    """
    test_gauth_path = f"{cred_dir}/{_gauth_file}"
    test_mqtt_path = f"{cred_dir}/{_mqtt_token_file}"
    if os.path.exists(test_mqtt_path):
        with open(test_mqtt_path, "r", encoding="utf8") as f:
            mqtt_json = f.read()
        try:
            mqtt_token = json.loads(mqtt_json)
        except json.decoder.JSONDecodeError as err:
            print(f"{err}, {test_mqtt_path}")
            os.remove(test_mqtt_path)
            return
        try:
            mqtt_claims = jwt.decode(mqtt_token["token"], options={"verify_signature": False})
        except Exception as err:
            print(f"{err}, {test_mqtt_path}")
            os.remove(test_mqtt_path)
            return

        exp = float(mqtt_claims["exp"])
        now = time.time()
        if expire and now < exp:
            return  # exit if expire request is still good
        # otherwise remove
        os.remove(test_mqtt_path)
    if os.path.exists(test_gauth_path):
        os.remove(test_gauth_path)


if __name__ == "__main__":
    globals()[sys.argv[1]]()
